from __future__ import division

import functools
from GangaCore.Utility.logging import getLogger

from GangaCore.Core.exceptions import (GangaException,
                                   InaccessibleObjectError,
                                   RepositoryError)

import time
import threading

from GangaCore.Core.GangaThread.GangaThread import GangaThread
from GangaCore.GPIDev.Lib.GangaList.GangaList import GangaList
from GangaCore.GPIDev.Base.Objects import GangaObject
from GangaCore.GPIDev.Schema import Schema, Version
from GangaCore.GPIDev.Base.Proxy import isType, getName
from GangaCore.Utility.Config import getConfig

logger = getLogger()

class RegistryError(GangaException):

    def __init__(self, what=''):
        super(RegistryError, self).__init__(self, what)
        self.what = what

    def __str__(self):
        return "RegistryError: %s" % self.what


class RegistryAccessError(RegistryError):

    """ This error is raised if the request is valid in principle, 
        but the Registry cannot be accessed at the moment."""

    def __init__(self, what=''):
        super(RegistryAccessError, self).__init__(what)

    def __str__(self):
        return "RegistryAccessError: %s" % self.what


class RegistryLockError(RegistryError):

    """ This error is raised if the request is valid in principle,
        but the object is locked by another Ganga session"""

    def __init__(self, what=''):
        super(RegistryLockError, self).__init__(what)

    def __str__(self):
        return "RegistryLockError: %s" % self.what


class ObjectNotInRegistryError(RegistryError):

    """ This error is raised if an object has been associated to this registry,
        but is not actually in the registry. This most probably indicates an internal Ganga error."""

    def __init__(self, what=''):
        super(ObjectNotInRegistryError, self).__init__(what)

    def __str__(self):
        return "ObjectNotInRegistryError: %s" % self.what


class RegistryKeyError(RegistryError, KeyError):

    """ This error is raised if the given id is not found in the registry """

    def __init__(self, what=''):
        super(RegistryKeyError, self).__init__(what)

    def __str__(self):
        return "RegistryKeyError: %s" % self.what


class RegistryIndexError(RegistryError, IndexError):

    """ This error is raised if the given id is not found in the registry """

    def __init__(self, what=''):
        super(RegistryIndexError, self).__init__(what)

    def __str__(self):
        return "RegistryIndexError: %s" % self.what


def makeRepository(registry):
    """Factory that selects, imports and instantiates the correct GangaRepository
    Args:
        registry (Registry): This maps the Registry type to the correct Repository
    """
    if registry.type in ["LocalXML", "LocalPickle"]:
        from GangaCore.Core.GangaRepository.GangaRepositoryXML import GangaRepositoryLocal
        return GangaRepositoryLocal(registry)
    elif registry.type in ["SQLite"]:
        from GangaCore.Core.GangaRepository.GangaRepositorySQLite import GangaRepositorySQLite
        return GangaRepositorySQLite(registry)
    elif registry.type in ["Transient"]:
        from GangaCore.Core.GangaRepository.GangaRepository import GangaRepository
        return GangaRepository(registry)
    elif registry.type in ["ImmutableTransient"]:
        from GangaCore.Core.GangaRepository.GangaRepositoryImmutableTransient import GangaRepositoryImmutableTransient
        return GangaRepositoryImmutableTransient(registry, registry.location, registry.file_ext, registry.pickle_files)
    else:
        raise RegistryError("Repository %s: Unknown repository type %s" % (registry.name, registry.type))


class IncompleteObject(GangaObject):

    """ This class represents an object that could not be loaded on startup"""

    _schema = Schema(Version(0, 0), {})
    _name = "IncompleteObject"
    _category = "internal"
    _hidden = 1

    _exportmethods = ['reload', 'remove', '__repr__']

    _additional_slots = ('registry', 'id')

    def __init__(self, registry, this_id):
        """
        This constructs an object which is placed into the objects dict when a repo fails to load an object due to some error
        Args:
            registry (Registry): This is the registry the object belongs to
            this_id (int): This is the registry/repo id of the object in the objects dict
        """
        super(IncompleteObject, self).__init__()
        self.registry = registry
        self.id = this_id

    def reload(self):
        """
        This will trigger a re-load of the object from disk which is useful if the object was locked but accessible by Ganga
        TODO work ouf if this is still called anywhere
        """
        with self.registry._flush_lock:
            with self.registry._read_lock:
                self.registry._load(self)
                logger.debug("Successfully reloaded '%s' object #%i!" % (self.registry.name, self.id))

    def remove(self):
        """
        This will trigger a delete of the the object itself from within the given Repository but not registry
        TODO work out if this is safe and still called
        """
        with self.registry._flush_lock:
            with self.registry._read_lock:
                if len(self.registry.repository.lock([self.id])) == 0:
                    errstr = "Could not lock '%s' object #%i!" % (self.registry.name, self.id)
                    try:
                        errstr += " Object is locked by session '%s' " % self.registry.repository.get_lock_session(self.id)
                    except Exception as err:
                        logger.debug("Remove Lock error: %s" % err)
                    raise RegistryLockError(errstr)
                self.registry.repository.delete([self.id])

    def __repr__(self):
        """
        This returns a repr of the object in question as inaccessible
        """
        return "Incomplete object in '%s', ID %i. Try reload() or remove()." % (self.registry.name, self.id)


def synchronised_flush_lock(f):
    """
    Specific flush lock as flushing can take a long time. This ensures no changes occur while flushing though reads can.

    Args:
        f (func): Function to decorate
    """
    @functools.wraps(f)
    def decorated(self, *args, **kwargs):
        with self._flush_lock:
            return f(self, *args, **kwargs)
    return decorated

def synchronised_read_lock(f):
    """
    General read lock for quick functions that won't take a while but won't change anything either

    Args:
        f (func): Function to decorate
    """
    @functools.wraps(f)
    def decorated(self, *args, **kwargs):
        with self._read_lock:
            return f(self, *args, **kwargs)
    return decorated

def synchronised_complete_lock(f):
    """
    Entirely lock the registry. Make sure locks are acquired in the right order!

    Args:
        f (func): Function to decorate
    """
    @functools.wraps(f)
    def decorated(self, *args, **kwargs):
        with self._flush_lock:
            with self._read_lock:
                return f(self, *args, **kwargs)
    return decorated

class RegistryFlusher(GangaThread):
    """
    This class is intended to be used by the registry to perfom
    automatic flushes on a fixed schedule so that information is not
    lost if Ganga is shut down abruptly.
    """

    __slots__ = ('registry', '_stop')

    def __init__(self, registry, *args, **kwargs):
        """
        This inits the RegistryFlusher and makes use of threading events
        Args:
            registry (Registry): The registry this Registry Flusher is flushing
            args (list): Args passed to the constructor of a new RegistryFlusher thread
            kwargs (dict): Kwds passed to the constructor of a new RegistryFlusher thread
        """
        super(RegistryFlusher, self).__init__(*args, **kwargs)
        self.registry = registry
        self._stop = threading.Event()

    def stop(self):
        """
        Ask the thread to stop what it is doing and it will finish
        the next chance it gets.
        TODO, does this need to be exposed as a method if only used internally?
        """
        self._stop.set()

    @property
    def stopped(self):
        """
        Returns if the flusher has stopped as a boolean?
        """
        return self._stop.isSet()

    def join(self, *args, **kwargs):
        """
        Called on thread shutdown to stop the active thread
        Args:
            args (list): Args passed to the constructor of a new RegistryFlusher thread
            kwargs (dict) : Kwds passed to the constructor of a new RegistryFlusher thread
        """
        self.stop()
        super(RegistryFlusher, self).join(*args, **kwargs)

    def run(self):
        """
        This will run an indefinite loop which periodically checks
        whether it should stop. In between calls to ``flush_all`` it
        will wait for a fixed period of time.
        """
        sleeps_per_second = 10  # This changes the granularity of the sleep.
        regConf = getConfig('Registry')
        while not self.stopped:
            sleep_period = regConf['AutoFlusherWaitTime']
            for i in range(sleep_period*sleeps_per_second):
                time.sleep(1/sleeps_per_second)
                if self.stopped:
                    return
            # This will trigger a flush on all dirty objects in the repo,
            # It will lock all objects dirty as a result of the nature of the flush command
            logger.debug('Auto-flushing: %s', self.registry.name)
            if regConf['EnableAutoFlush']:
                self.registry.flush_all()
        logger.debug("Auto-Flusher shutting down for Registry: %s" % self.registry.name)


class Registry(object):

    """Ganga Registry
    Base class providing a dict-like locked and lazy-loading interface to a Ganga repository
    """

    __slots__ = ('name', 'doc', '_hasStarted', '_needs_metadata', 'metadata', '_read_lock', '_flush_lock', '_parent', 'repository', '_objects', '_incomplete_objects', 'flush_thread', 'type', 'location')

    def __init__(self, name, doc):
        """Registry constructor, giving public name and documentation
        Args:
            name (str): Name of teh registry e.g. 'jobs', 'box', 'prep'
            doc (str): This is the doc string of the registry describing it's use and contents
        """
        self.name = name
        self.doc = doc
        self._hasStarted = False
        self._needs_metadata = False
        self.metadata = None
        self._read_lock = threading.RLock()
        self._flush_lock = threading.RLock()

        self._parent = None

        self.repository = None
        self._objects = None
        self._incomplete_objects = None

        self.flush_thread = None

    def hasStarted(self):
        """
        Wrapper function to return _hasStarted boolen
        TODO is this needed over accessing the boolean, should the boolean be 'public'
        """
        return self._hasStarted

    # Methods intended to be called from ''outside code''
    def __getitem__(self, this_id):
        """ Returns the Ganga Object with the given id.
            Raise RegistryKeyError
        Args:
            this_id (int): This is the key of an object in the object dictionary
        """
        logger.debug("__getitem__")
        # Is this an Incomplete Object?
        if this_id in self._incomplete_objects:
            return IncompleteObject(self, this_id)

        # Nope, so try to find it and raise an exception if not
        try:
            return self._objects[this_id]
        except KeyError as err:
            logger.debug("Repo KeyError: %s" % err)
            logger.debug("Keys: %s id: %s" % (self._objects.keys(), this_id))
            raise RegistryKeyError("Could not find object #%s" % this_id)

    @synchronised_read_lock
    def __len__(self):
        """ Returns the current number of root objects """
        logger.debug("__len__")
        return len(self._objects)

    @synchronised_read_lock
    def __contains__(self, this_id):
        """ Returns True if the given ID is in the registry
        Args:
            this_id (int): This is the key of an object in the object dictionary
        """
        logger.debug("__contains__")
        return this_id in self._objects

    def updateLocksNow(self):
        """
        Update the registry locks now. This is explicitly called from a method which can cause Ganga to pause for a long time
        TODO: determine if this is still valid
        """
        logger.debug("updateLocksNow")
        self.repository.updateLocksNow()

    @synchronised_read_lock
    def ids(self):
        """ Returns the list of ids of this registry """
        logger.debug("ids")
        return sorted(self._objects.keys())

    @synchronised_read_lock
    def items(self):
        """ Return the items (ID,obj) in this registry. 
        Recommended access for iteration, since accessing by ID can fail if the ID iterator is old"""
        return sorted(self._objects.items())

    def iteritems(self):
        """ Return the items (ID,obj) in this registry."""
        logger.debug("iteritems")
        returnable = self.items()
        return returnable

    @synchronised_read_lock
    def keys(self):
        """ Returns the list of ids of this registry """
        logger.debug("keys")
        returnable = self.ids()
        return returnable

    @synchronised_read_lock
    def values(self):
        """ Return the objects in this registry, in order of ID.
        Besides items() this is also recommended for iteration."""
        logger.debug("values")
        returnable = [it[1] for it in self.items()]
        return returnable

    def __iter__(self):
        """ Return an iterator for the self.values list """
        logger.debug("__iter__")
        returnable = iter(self.values())
        return returnable

    def find(self, obj):
        """Returns the id of the given object in this registry, or 
        Raise ObjectNotInRegistryError if the Object is not found
        Args:
            _obj (GangaObject): This is the object we want to match in the objects repo
        """
        try:
            return next(id_ for id_, o in self._objects.items() if o is obj)
        except StopIteration:
            raise ObjectNotInRegistryError("Object '%s' does not seem to be in this registry: %s !" % (getName(obj), self.name))

    @synchronised_complete_lock
    def clean(self, force=False):
        """Deletes all elements of the registry, if no other sessions are present.
        if force == True it removes them regardless of other sessions.
        Returns True on success, False on failure.
        Args:
            force (bool): This will force a clean action to be performed even if there are other active Ganga sessions, EXTREMELY UNFRIENDLY!!!
        """
        logger.debug("clean")
        if self.hasStarted() is not True:
            raise RegistryAccessError("Cannot clean a disconnected repository!")

        if not force:
            other_sessions = self.repository.get_other_sessions()
            if len(other_sessions) > 0:
                logger.error("The following other sessions are active and have blocked the clearing of the repository: \n * %s" % ("\n * ".join(other_sessions)))
                return False
        self.repository.reap_locks()
        self.repository.delete(self._objects.keys())
        self.repository.clean()

    # Methods that can be called by derived classes or Ganga-internal classes like Job
    # if the dirty objects list is modified, the methods must be locked by self._lock
    # all accesses to the repository must also be locked!

    @synchronised_complete_lock
    def _add(self, obj, force_index=None):
        """ Add an object to the registry and assigns an ID to it. 
        use force_index to set the index (for example for metadata). This overwrites existing objects!
        Raises RepositoryError
        Args:
            _obj (GangaObject): This is th object which is to be added to this Registy/Repo
            force_index (int): This is the index which we will give the object, None = auto-assign
        """
        logger.debug("_add")

        if self.hasStarted() is not True:
            raise RepositoryError("Cannot add objects to a disconnected repository!")

        if force_index is None:
            ids = self.repository.add([obj])
        else:
            if len(self.repository.lock([force_index])) == 0:
                raise RegistryLockError("Could not lock '%s' id #%i for a new object!" % (self.name, force_index))
            ids = self.repository.add([obj], [force_index])

        obj._setRegistry(self)
        obj._registry_locked = True

        self.repository.flush(ids)

        return ids[0]

    @synchronised_complete_lock
    def _remove(self, obj, auto_removed=0):
        """ Private method removing the obj from the registry. This method always called.
        This method may be overriden in the subclass to trigger additional actions on the removal.
        'auto_removed' is set to true if this method is called in the context of obj.remove() method to avoid recursion.
        Only then the removal takes place. In the opposite case the obj.remove() is called first which eventually calls
        this method again with "auto_removed" set to true. This is done so that obj.remove() is ALWAYS called once independent
        on the removing context.
        Raise RepositoryError
        Raise RegistryAccessError
        Raise RegistryLockError
        Raise ObjectNotInRegistryError
        Args:
            _obj (GangaObject): The object which we want to remove from the Repo/Registry
            auto_removed (int, bool): True/False for if the object can be auto-removed by the base Repository method
        """
        logger.debug("_remove")

        if self.hasStarted() is not True:
            raise RegistryAccessError("Cannot remove objects from a disconnected repository!")

        if not auto_removed and hasattr(obj, "remove"):
            obj.remove()
        else:
            this_id = self.find(obj)
            self._acquire_session_lock(obj)

            logger.debug('deleting the object %d from the registry %s', this_id, self.name)
            self.repository.delete([this_id])

    @synchronised_flush_lock
    def _flush(self, objs):
        """
        Flush a set of objects to the persistency layer immediately

        Only those objects passed in will be flushed and only if they are dirty.

        Args:
            objs (list): a list of objects to flush
        """
        # Too noisy
        #logger.debug("_flush")

        if not isType(objs, (list, tuple, GangaList)):
            objs = [objs]

        if self.hasStarted() is not True:
            raise RegistryAccessError("Cannot flush to a disconnected repository!")

        for obj in objs:
            # check if the object is dirty, if not do nothing
            if not obj._dirty:
                continue

            if not self.has_loaded(obj):
                continue

            with obj.const_lock:
                # flush the object
                obj_id = self.find(obj)
                self.repository.flush([obj_id])
                obj._setFlushed()

    def flush_all(self):
        """
        This will attempt to flush all the jobs in the registry.
        It does this via ``_flush`` so the same conditions apply.
        """
        if self.hasStarted():
            for _obj in self.values():
                self._flush(_obj)

        if self.metadata and self.metadata.hasStarted():
            self.metadata.flush_all()

    def _load(self, obj):
        """
        Use this function to load an object from disk as it will check if the object is already loaded *outside*
         of the lock for when the actual load takes place in _locked_load
        """
        if not self.repository.isObjectLoaded(obj):
            self._locked_load(obj)

    @synchronised_complete_lock
    def _locked_load(self, obj):
        """
        Fully load an object from a Repo/disk into memory. Should only ever be called from _load!
        Args:
            obj (GangaObject): This is the object we want to fully load
        """
        logger.debug("_locked_load")

        # find the object ID
        obj_id = self.find(obj)

        try:
            self.repository.load([obj_id])
        except Exception as err:
            logger.error("Error Loading Job! '%s'" % obj_id)
            # Cleanup after ourselves if an error occured
            # Didn't load mark as clean so it's not flushed
            if obj_id in self._objects:
                self._objects[obj_id]._setFlushed(auto_load_deps=False)
            raise

    def _acquire_session_lock(self, obj):
        """Obtain write access on a given object.
        Raise RepositoryError
        Raise RegistryAccessError
        Raise RegistryLockError
        Raise ObjectNotInRegistryError (via self.find())
        Args:
            _obj (GangaObject): This is the object we want to get write access to and lock on disk
        """
        if self.hasStarted() is not True:
            raise RegistryAccessError("Cannot get write access to a disconnected repository!")

        if not hasattr(obj, '_registry_locked') or not obj._registry_locked:
            this_id = self.find(obj)
            if len(self.repository.lock([this_id])) == 0:
                errstr = "Could not lock '%s' object #%i!" % (self.name, this_id)
                errstr += " Object is locked by session '%s' " % self.repository.get_lock_session(this_id)
                raise RegistryLockError(errstr)

            obj._registry_locked = True

    def _release_session_lock_and_flush(self, obj):
        """Release the lock on a given object.
        Raise RepositoryError
        Raise RegistryAccessError
        Raise ObjectNotInRegistryError
        Args:
            obj (GangaObject): This is the object we want to release the file lock for
        """
        if self.hasStarted() is not True:
            raise RegistryAccessError("Cannot manipulate locks of a disconnected repository!")
        logger.debug("Reg: %s _release_lock(%s)" % (self.name, self.find(obj)))
        if hasattr(obj, '_registry_locked') and obj._registry_locked:
            oid = self.find(obj)
            self.repository.flush([oid])
            obj._registry_locked = False
            self.repository.unlock([oid])

    def getIndexCache(self, obj):
        """Returns a dictionary to be put into obj._index_cache (is this valid)
        This can and should be overwritten by derived Registries to provide more index values."""
        return {}

    @synchronised_complete_lock
    def startup(self):
        """Connect the repository to the registry. Called from Repository_runtime.py"""
        try:
            self._hasStarted = True
            t0 = time.time()
            self.repository = makeRepository(self)
            self._objects = self.repository.objects
            self._incomplete_objects = self.repository.incomplete_objects

            if self._needs_metadata:
                t2 = time.time()
                if self.metadata is None:
                    self.metadata = Registry(self.name + ".metadata", "Metadata repository for %s" % self.name)
                    self.metadata.type = self.type
                    self.metadata.location = self.location
                    setattr(self.metadata, '_parent', self) ## rcurrie Registry has NO '_parent' Object so don't understand this is this used for JobTree?
                logger.debug("metadata startup")
                self.metadata.startup()
                t3 = time.time()
                logger.debug("Startup of %s.metadata took %s sec" % (self.name, t3-t2))

            logger.debug("repo startup")
            #self.hasStarted() = True
            self.repository.startup()
            t1 = time.time()
            logger.debug("Registry '%s' [%s] startup time: %s sec" % (self.name, self.type, t1 - t0))
        except Exception as err:
            logger.debug("Logging Repo startup Error: %s" % err)
            self._hasStarted = False
            raise
        #finally:
        #    pass

        # Now we check the repo for faults and for inconsistent objects
        reg_config = getConfig('Registry')

        if reg_config['DisableLoadCheck']:
            return
        else:
            self.check()

    def check(self):
        """ This stub allows for a Registry class to perform maintenance on startup after the repo has been initialized
        This is intended to be called after the startup proceedure has been finished """
        pass

    @synchronised_complete_lock
    def shutdown(self):
        """Flush and disconnect the repository. Called from Repository_runtime.py """
        from GangaCore.Utility.logging import getLogger
        logger = getLogger()
        logger.debug("Shutting Down Registry")
        logger.debug("shutdown")
        try:
            self._hasStarted = True
            # NB flush_all by definition relies on both the metadata repo and the repo to be fully initialized
            try:
                self.flush_all()
            except Exception as err:
                logger.error("Exception on flushing '%s' registry: %s", self.name, err)

            # Now we can safely shutdown the metadata repo if one is loaded
            try:
                if self.metadata is not None:
                    self.metadata.shutdown()
            except Exception as err:
                logger.debug("Exception on shutting down metadata repository '%s' registry: %s", self.name, err)
                raise

            # Now we can release locks on the objects we have
            for obj in self._objects.values():
                # locks are not guaranteed to survive repository shutdown
                obj._registry_locked = False
            self.repository.shutdown()

            self.metadata = None

        finally:
            self._hasStarted = False

    def info(self, full=False):
        """Returns an informative string onFlush and disconnect the repository. Called from Repository_runtime.py """
        logger.debug("info")
        s = "registry '%s': %i objects" % (self.name, len(self._objects))
        if full:
            other_sessions = self.repository.get_other_sessions()
            if len(other_sessions) > 0:
                s += ", %i other concurrent sessions:\n * %s" % (len(other_sessions), "\n * ".join(other_sessions))
        return s

    def has_loaded(self, obj):
        """Returns True/False for if a given object has been fully loaded by the Registry.
        Returns False on the object not being in the Registry!
        This ONLY applies to master jobs as the Registry has no apriori knowledge of the subjob structure.
        Consult SubJobXMLList for a more fine grained loaded/not-loaded info/test.
        Args:
            obj (GangaObject): Object which we want to look for in this repo
        """
        try:
            index = self.find(obj)
        except ObjectNotInRegistryError:
            return False

        return self.repository.isObjectLoaded(obj)




################################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: PrimeFactorizer.py,v 1.2 2009-04-23 13:47:59 moscicki Exp $
################################################################################

from GangaCore.GPIDev.Adapters.IApplication import IApplication
from GangaCore.GPIDev.Adapters.IRuntimeHandler import IRuntimeHandler
from GangaCore.GPIDev.Schema import *
from GangaCore.Utility.Config import getConfig
from GangaCore.GPIDev.Lib.File import File

class PrimeFactorizer(IApplication):
    """
    PrimeFactorizer application -- factorize any arbitrary number into prime numbers.
    """
    _schema = Schema(Version(1,1), {
        'number' : SimpleItem(defvalue=1,sequence=0,typelist=['int','long'],doc='The number to be factorized.'),
        } )

    _category = 'applications'
    _name = 'PrimeFactorizer'
    #_GUIPrefs = [ { 'attribute' : 'number', 'widget' : 'Integer' },
    #              { 'attribute' : 'prime_table', 'widget' : 'String' }]

    def __init__(self):
        super(PrimeFactorizer,self).__init__()

    def configure(self,masterappconfig):
        from GangaCore.GPI import TUTDIR
        from GangaCore.Core.exceptions import ApplicationConfigurationError
        import os.path

        ## the prime number factorizer executable
        self.exe = File(TUTDIR + '/Lib/primes/prime_factor.py') 

        prime_tables = []
        try:
            job = self.getJobObject()
            prime_tables = job.inputdata.get_dataset()
        except:
            pass

        ## the arguments of the executable
        ##  - arg1: the number
        ##  - arg2: the SURL of the prime number lookup table
        self.args    = ['%s' % self.number] + prime_tables
        self.inputs  = []
        self.outputs = ['*.dat']
        self.envs = {} 
        return (None,None)

# FIXME: a cleaner solution, which is integrated with type information in schemas should be used automatically
#config = getConfig('GangaTutorial_Properties')
mc = getConfig('MonitoringServices')
mc.addOption('GangaTutorial',None,'')

class RTHandler(IRuntimeHandler):
    def prepare(self,app,appconfig,appmasterconfig,jobmasterconfig):
        from GangaCore.GPIDev.Adapters.StandardJobConfig import StandardJobConfig

        c = StandardJobConfig(app.exe,app.inputs,app.args,app.outputs,app.envs)
        c.monitoring_svc = mc['GangaTutorial']

        return c

class LCGRTHandler(IRuntimeHandler):
    def prepare(self,app,appconfig,appmasterconfig,jobmasterconfig):
        from GangaCore.Lib.LCG import LCGJobConfig

        c = LCGJobConfig(app.exe,app.inputs,app.args,app.outputs,app.envs)
        c.monitoring_svc = mc['GangaTutorial']

        return c

from GangaCore.GPIDev.Adapters.ApplicationRuntimeHandlers import allHandlers

allHandlers.add('PrimeFactorizer','LSF', RTHandler)
allHandlers.add('PrimeFactorizer','Local', RTHandler)
allHandlers.add('PrimeFactorizer','PBS', RTHandler)
allHandlers.add('PrimeFactorizer','SGE', RTHandler)
allHandlers.add('PrimeFactorizer','Condor', RTHandler)
allHandlers.add('PrimeFactorizer','LCG', LCGRTHandler)
allHandlers.add('PrimeFactorizer','TestSubmitter', RTHandler)
allHandlers.add('PrimeFactorizer','Interactive', RTHandler)
allHandlers.add('PrimeFactorizer','Batch', RTHandler)


##############################################################
## handler for NG

from GangaCore.GPIDev.Adapters.ApplicationRuntimeHandlers import allHandlers
from GangaCore.GPIDev.Adapters.IRuntimeHandler import IRuntimeHandler

class NGRTHandler(IRuntimeHandler):

    def prepare(self,app,appconfig,appmasterconfig,jobmasterconfig):
        from GangaNG.Lib.NG import NGJobConfig
        return NGJobConfig(app.exe,app.inputs,app.args,app.outputs,app.envs)



from __future__ import print_function

#from GangaCore.Lib.LCG.GridSimulator import GridSimulator
#g = GridSimulator()
#gridmap_filename = g.gridmap_filename
#finished_jobs_filename = g.finished_jobs_filename

import sys
import os.path
import shelve


try:
    basedir = sys.argv[1]
except IndexError:
    basedir = '.'

gridmap_filename = "%s/lcg_simulator_gridmap" % basedir
finished_jobs_filename = "%s/lcg_simulator_finished_jobs" % basedir


gridmap = shelve.open(gridmap_filename)
finished_jobs = shelve.open(finished_jobs_filename)

deltas = []
job_finished_times = []
ganga_finished_times = []

for gid in gridmap:
    if gid[0] == '_':
        continue
    params = eval(file(os.path.join(gridmap[gid], 'params')).read())
    try:
        job_finished_times.append(params['expected_finish_time'])
        ganga_finished_times.append(finished_jobs[gid])
        deltas.append(ganga_finished_times[-1] - job_finished_times[-1])
    except KeyError:
        print('Missing data for:', gid, file=sys.stderr)

idle_cnt = 0
idle = []

job_finished_times.sort()
ganga_finished_times.sort()

i = 0
j = 0

INF = 1e40

start_t = min(job_finished_times[0], ganga_finished_times[0])

while i < len(job_finished_times) and j < len(ganga_finished_times):

    try:
        a = job_finished_times[i]
    except IndexError:
        a = INF

    try:
        b = ganga_finished_times[j]
    except IndexError:
        b = INF

    if a < b:
        idle_cnt += 1
        idle.append((a - start_t, idle_cnt))
        i += 1
    elif a > b:
        idle_cnt -= 1
        idle.append((b - start_t, idle_cnt))
        j += 1
    else:
        idle.append((a - start_t, idle_cnt))
        i += 1
        j += 1


f = file('%s/idle.dat' % basedir, 'w')
f.write(
    "# time-based counter of jobs which were reported by the grid as finished but not completed/failed in ganga\n")
f.write(
    "# x = time in seconds from the beginning of the analysis, y = counter of 'idle' jobs\n")
for i in idle:
    f.write("%d %d\n" % i)

f = file('%s/deltas.dat' % basedir, 'w')
f.write(
    "# time difference (for each individual job) between the job was reported by the grid as finished and completed/failed in ganga\n")
for d in deltas:
    f.write(d + '\n')



import os
import shutil
from datetime import datetime
import tempfile
import tarfile
import random
import threading
import uuid
import shutil
from GangaCore.Core.exceptions import ApplicationConfigurationError, ApplicationPrepareError, GangaException, GangaFileError
from GangaCore.GPIDev.Adapters.ApplicationRuntimeHandlers import allHandlers
from GangaCore.GPIDev.Adapters.IRuntimeHandler import IRuntimeHandler
from GangaCore.GPIDev.Adapters.StandardJobConfig import StandardJobConfig
from GangaCore.GPIDev.Base.Proxy import getName
from GangaCore.GPIDev.Lib.File.File import File, ShareDir
from GangaCore.GPIDev.Lib.File.FileBuffer import FileBuffer
from GangaCore.GPIDev.Lib.File.LocalFile import LocalFile
from GangaCore.GPIDev.Lib.File.OutputFileManager import getWNCodeForOutputPostprocessing
from GangaCore.Utility.Config import getConfig
from GangaCore.Utility.logging import getLogger
from GangaCore.Utility.util import unique
from GangaCore.GPIDev.Credentials.CredentialStore import credential_store

from GangaDirac.Lib.Files.DiracFile import DiracFile
from GangaDirac.Lib.RTHandlers.DiracRTHUtils import dirac_inputdata, dirac_ouputdata, mangle_job_name, diracAPI_script_settings, API_nullifier
from GangaDirac.Lib.Utilities.DiracUtilities import execute, GangaDiracError
from GangaGaudi.Lib.RTHandlers.RunTimeHandlerUtils import master_sandbox_prepare, sandbox_prepare, script_generator
from GangaLHCb.Lib.RTHandlers.RTHUtils import lhcbdiracAPI_script_template, lhcbdirac_outputfile_jdl
from GangaLHCb.Lib.LHCbDataset.LHCbDataset import LHCbDataset
from ..Applications.GaudiExecUtils import addTimestampFile
from GangaGaudi.Lib.Applications.GaudiUtils import gzipFile

logger = getLogger()

_pseudo_session_id = str(uuid.uuid4())

def genDataFiles(job):
    """
    Generating a data.py file which contains the data we want gaudirun to use
    Args:
        job (Job): This is the job object which contains everything useful for generating the code
    """
    logger.debug("Doing XML Catalog stuff")

    inputsandbox = []

    data = job.inputdata
    if data:
        logger.debug("Returning options String")
        data_str = data.optionsString()
        if data.hasLFNs():
            logger.info("Generating Data catalog for job: %s" % job.fqid)
            logger.debug("Returning Catalogue")
            inputsandbox.append(FileBuffer('catalog.xml', data.getCatalog()))
            cat_opts = '\nfrom Gaudi.Configuration import FileCatalog\nFileCatalog().Catalogs = ["xmlcatalog_file:catalog.xml"]\n'
            data_str += cat_opts

        inputsandbox.append(FileBuffer(GaudiExecDiracRTHandler.data_file, data_str))
    else:
        inputsandbox.append(FileBuffer(GaudiExecDiracRTHandler.data_file, '#dummy_data_file\n'+LHCbDataset().optionsString()))

    return inputsandbox


def getAutoDBTags(job):
    knownApps = ['DaVinci', 'Brunel', 'Moore']
    prefix = 'DaVinci'
    for app in knownApps:
        if app in job.application.directory:
            prefix = app
    inputsandbox = []
    ddb, conddb = execute('getDBtagsFromLFN("{0}")'.format(job.inputdata[0].lfn)) # take the tags only from the first file
    tagOpts = 'from Configurables import ' + prefix +'\n' 
    tagOpts += prefix + '().DDDBtag = ' + "'" + ddb + "'\n"
    tagOpts += prefix + '().CondDBtag = ' + "'" + conddb + "'"
    return tagOpts

def generateWrapperScript(app):
    """
    This generates the wrapper script which is run for non GaudiExec type apps
    Args:
        app (GaudiExec): GaudiExec instance which contains the script to run on the WN
    """

    return FileBuffer(name=app.getWrapperScriptName(), contents=app.getWNPythonContents())


def getScriptName(app):
    """
    Returns the name of the script which runs on the WN.
    Args:
        app (Job): This is the app object which contains everything useful for generating the code
    """
    job = app.getJobObject()
    return "_".join((getConfig('Configuration')['user'], getName(app), 'Job', job.getFQID('.'), _pseudo_session_id, 'script'))+'.py'


def generateWNScript(commandline, app):
    """
    Generate the script as a file buffer and return it
    Args:
        commandline (str): This is the command-line argument the script is wrapping
        app (Job): This is the app object which contains everything useful for generating the code
    """
    job = app.getJobObject()
    exe_script_name = getScriptName(app)

    return FileBuffer(name=exe_script_name, contents=script_generator(gaudiRun_script_template(), COMMAND=commandline,
                                                                      OUTPUTFILESINJECTEDCODE = getWNCodeForOutputPostprocessing(job, '    ')),
                      subdir='jobScript', executable=True)


def collectPreparedFiles(app):
    """
    Collect the files from the Application in the prepared state
    Args:
        app (GaudiExec): This expects only the GaudiExec app
    """
    if not isinstance(app.is_prepared, ShareDir):
        raise ApplicationConfigurationError('Failed to prepare Application Correctly')
    shared_dir = app.getSharedPath()
    input_files, input_folders = [], []
    for root, dirs, files in os.walk(shared_dir, topdown=True):
        for name in files:
            input_files.append(os.path.join(root, name))
        for name in dirs:
            input_folders.append(os.path.join(root, name))

    for file_ in app.getJobObject().inputfiles:
        if isinstance(file_, LocalFile):
            shutil.copy(os.path.join(file_.localDir, os.path.basename(file_.namePattern)), shared_dir)
            input_files.append(os.path.join(shared_dir, file_.namePattern))
        elif not isinstance(file_, DiracFile):
            raise ApplicationConfigurationError(None, "File type: %s Not _yet_ supported in GaudiExec" % type(file_))

    return input_files, input_folders


def prepareCommand(app):
    """
    Returns the command which is to be run on the worker node
    Args:
        app (GaudiExec): This expects only the GaudiExec app
    """

    all_opts_files = app.getOptsFiles()
    opts_names = []
    for opts_file in all_opts_files:
        if isinstance(opts_file, (LocalFile, DiracFile)):
            # Ideally this would NOT need the basename, however LocalFile is special in this regard.
            # TODO Fix this after fixing LocalFile
            opts_names.append(os.path.basename(opts_file.namePattern))
        else:
            raise ApplicationConfigurationError("The filetype: %s is not yet supported for use as an opts file.\nPlease contact the Ganga devs is you wish this implemented." %
                                                getName(opts_file))

    sourceEnv = app.getWNEnvScript()

    run_cmd = ' export ganga_jobid=%s && ./run ' % app.getJobObject().fqid

    if not app.useGaudiRun:
        full_cmd = sourceEnv + run_cmd + 'python %s' % app.getWrapperScriptName()
    else:
        #If the job does not have inputdata don't include the data.py file in the run command. For Gauss jobs.
        if app.getJobObject().inputdata:
            full_cmd = sourceEnv + run_cmd + "gaudirun.py %s %s" % (' '.join(opts_names), GaudiExecDiracRTHandler.data_file)
        else:
            full_cmd = sourceEnv + run_cmd + "gaudirun.py %s " % (' '.join(opts_names))
        if app.extraOpts:
            full_cmd += ' ' + app.getExtraOptsFileName()
        if app.getMetadata:
            full_cmd += ' summary.py'
        if app.autoDBtags:
            full_cmd += ' dbTags.py'
        if app.extraArgs:
            full_cmd += " " + " ".join(app.extraArgs)

    return full_cmd


class GaudiExecRTHandler(IRuntimeHandler):

    """The runtime handler to run plain executables on the Local backend"""

    def master_prepare(self, app, appmasterconfig):
        """
        Prepare the RTHandler for the master job so that applications to be submitted
        Args:
            app (GaudiExec): This application is only expected to handle GaudiExec Applications here
            appmasterconfig (unknown): Output passed from the application master configuration call
        """
        if app.autoDBtags and not app.getJobObject().inputdata[0].lfn.startswith('/lhcb/MC/'):
            logger.warning("This doesn't look like MC! Not automatically adding db tags.")
            app.autoDBtags = False

        inputsandbox, outputsandbox = master_sandbox_prepare(app, appmasterconfig)

        if isinstance(app.jobScriptArchive, LocalFile):
            app.jobScriptArchive = None

        generateJobScripts(app, appendJobScripts=True)

        scriptArchive = os.path.join(app.jobScriptArchive.localDir, app.jobScriptArchive.namePattern)

        inputsandbox.append(File(name=scriptArchive))

        if app.getMetadata:
            logger.info("Adding options to make the summary.xml")
            inputsandbox.append(FileBuffer('summary.py', "\nfrom Gaudi.Configuration import *\nfrom Configurables import LHCbApp\nLHCbApp().XMLSummary='summary.xml'"))

        if app.autoDBtags:
            logger.info("Adding options for auto DB tags")
            inputsandbox.append(FileBuffer('dbTags.py', getAutoDBTags(app.getJobObject())))

        return StandardJobConfig(inputbox=unique(inputsandbox), outputbox=unique(outputsandbox))

    def prepare(self, app, appconfig, appmasterconfig, jobmasterconfig):
        """
        Prepare the job in order to submit to the Local backend
        Args:
            app (GaudiExec): This application is only expected to handle GaudiExec Applications here
            appconfig (unknown): Output passed from the application configuration call
            appmasterconfig (unknown): Output passed from the application master_configure call
            jobmasterconfig (tuple): Output from the master job prepare step
        """

        job = app.getJobObject()

        # Setup the command which to be run and the input and output
        input_sand = genDataFiles(job)
        output_sand = []

        # If we are getting the metadata we need to make sure the summary.xml is added to the output sandbox if not there already.
        if app.getMetadata and not 'summary.xml' in output_sand:
            output_sand += ['summary.xml']

        # NB with inputfiles the mechanics of getting the inputfiled to the input of the Localhost backend is taken care of for us
        # We don't have to do anything to get our files when we start running
        # Also we don't manage the outputfiles here!

        job_command = prepareCommand(app)

        # Generate the script which is to be executed for us on the WN
        scriptToRun = generateWNScript(job_command, app)
        input_sand.append(scriptToRun)

        logger.debug("input_sand: %s" % input_sand)

        # It's this authors opinion that the script should be in the PATH on the WN
        # As it stands policy is that is isn't so we have to call it in a relative way, hence "./"
        c = StandardJobConfig('./'+os.path.join(scriptToRun.subdir, scriptToRun.name), input_sand, [], output_sand)
        return c


allHandlers.add('GaudiExec', 'Local', GaudiExecRTHandler)
allHandlers.add('GaudiExec', 'Condor', GaudiExecRTHandler)
allHandlers.add('GaudiExec', 'Interactive', GaudiExecRTHandler)
allHandlers.add('GaudiExec', 'Batch', GaudiExecRTHandler)
allHandlers.add('GaudiExec', 'LSF', GaudiExecRTHandler)
allHandlers.add('GaudiExec', 'PBS', GaudiExecRTHandler)
allHandlers.add('GaudiExec', 'SGE', GaudiExecRTHandler)


def generateDiracInput(app):
    """
    Construct a DIRAC input which does not need to be unique to each job but is required to have a unique checksum.
    This generates a unique file, uploads it to DRIAC and then stores the LFN in app.uploadedInput
    Args:
        app (GaudiExec): This expects a GaudiExec app to be passed so that the constructed
    """

    input_files, input_folders = collectPreparedFiles(app)

    job = app.getJobObject()

    if input_folders:
        raise ApplicationConfigurationError('Prepared folders not supported yet, please fix this in future')
    else:
        prep_dir = app.getSharedPath()
        addTimestampFile(prep_dir)
        master_no = job.id
        prep_file = '%s_%s.tgz' % (master_no, _pseudo_session_id)
        tmp_dir = tempfile.gettempdir()
        compressed_file = os.path.join(tmp_dir, 'diracInputFiles_'+os.path.basename(prep_file))

        if not job.master:
            rjobs = job.subjobs
        else:
            rjobs = [job]

        with tarfile.open(compressed_file, "w:gz") as tar_file:
            for name in input_files:
                # FIXME Add support for subfiles here once it's working across multiple IGangaFile objects in a consistent way
                # Not hacking this in for now just in-case we end up with a mess as a result
                tar_file.add(name, arcname=os.path.basename(name))

    new_df = uploadLocalFile(job, os.path.basename(compressed_file), tmp_dir)

    app.uploadedInput = new_df
    app.is_prepared.addAssociatedFile(DiracFile(lfn = new_df.lfn))

def generateJobScripts(app, appendJobScripts):
    """
    Construct a DIRAC scripts which must be unique to each job to have unique checksum.
    This generates a unique file, uploads it to DRIAC and then stores the LFN in app.uploadedInput
    Args:
        app (GaudiExec): This expects a GaudiExec app to be passed so that the constructed
        appendJobScripts (bool): Should we add the job scripts to the script archive? (Only makes sense on backends which auto-extact tarballs before running)
    """

    job = app.getJobObject()

    if not job.master:
        rjobs = job.subjobs or [job]
    else:
        rjobs = [job]

    tmp_dir = tempfile.gettempdir()

    # First create the extraOpts files needed 1 per subjob
    for this_job in rjobs:
        logger.debug("RTHandler Making Scripts: %s" % this_job.fqid)
        this_job.application.constructExtraFiles(this_job)

    if not job.master and job.subjobs:
        for sj in rjobs:
            sj.application.jobScriptArchive = sj.master.application.jobScriptArchive

    master_job = job.master or job

    # Now lets get the name of this tar file
    scriptArchive = os.path.join(master_job.application.jobScriptArchive.localDir, master_job.application.jobScriptArchive.namePattern)

    if appendJobScripts:
        # Now lets add the Job scripts to this archive and potentially the extra options to generate the summary.xml
        with tarfile.open(scriptArchive, 'a') as tar_file:
            if app.getMetadata:
                summaryScript = "\nfrom Gaudi.Configuration import *\nfrom Configurables import LHCbApp\nLHCbApp().XMLSummary='summary.xml'"
                summaryPath = os.path.join(job.getInputWorkspace().getPath(), 'summary.py')
                summaryFile = FileBuffer(summaryPath, summaryScript)
                summaryFile.create()
                tar_file.add(summaryPath, arcname = 'summary.py')
            if app.autoDBtags:
                dbScript = getAutoDBTags(job)
                dbPath = os.path.join(job.getInputWorkspace().getPath(), 'dbTags.py')
                dbFile = FileBuffer(dbPath, dbScript)
                dbFile.create()
                tar_file.add(dbPath, arcname = 'dbTags.py')
            for this_job in rjobs:
                this_app = this_job.application
                wnScript = generateWNScript(prepareCommand(this_app), this_app)
                this_script = os.path.join(tmp_dir, wnScript.name)
                wnScript.create(this_script)
                tar_file.add(this_script, arcname=os.path.join(wnScript.subdir, wnScript.name))
                os.unlink(this_script)

    gzipFile(scriptArchive, scriptArchive+'.gz', True)
    app.jobScriptArchive.namePattern = app.jobScriptArchive.namePattern + '.gz'

def generateDiracScripts(app):
    """
    Construct a DIRAC scripts which must be unique to each job to have unique checksum.
    This generates a unique file, uploads it to DRIAC and then stores the LFN in app.uploadedInput
    Args:
        app (GaudiExec): This expects a GaudiExec app to be passed so that the constructed
    """
    generateJobScripts(app, appendJobScripts=True)

    job = app.getJobObject()
    new_df = uploadLocalFile(job, app.jobScriptArchive.namePattern, app.jobScriptArchive.localDir)

    app.jobScriptArchive = new_df

    app.is_prepared.addAssociatedFile(DiracFile(lfn=new_df.lfn))

def uploadLocalFile(job, namePattern, localDir, should_del=True):
    """
    Upload a locally available file to the grid as a DiracFile.
    Randomly chooses an SE.

    Args:
        namePattern (str): name of the file
        localDir (str): localDir of the file
        should_del = (bool): should we delete the local file?
    Return
        DiracFile: a DiracFile of the uploaded LFN on the grid
    """

    new_df = DiracFile(namePattern, localDir=localDir)
    new_df.credential_requirements=job.backend.credential_requirements
    trySEs = getConfig('DIRAC')['allDiracSE']
    random.shuffle(trySEs)
    new_lfn = os.path.join(getInputFileDir(job), namePattern)
    returnable = None
    for SE in trySEs:
        #Check that the SE is writable
        if execute('checkSEStatus("%s", "%s")' % (SE, 'Write')):
            try:
                returnable = new_df.put(force=True, uploadSE=SE, lfn=new_lfn)[0]
                #We have to check the failureReason as DiracFile put doesn't necessarily raise an exception on failure
                if not returnable.failureReason=='':
                    logger.warning("Upload of input file as LFN %s to SE %s failed, trying another SE" % (new_lfn, SE))
                    #Clear the failure reason and continue
                    new_df.failureReason = ''
                    continue
                else:
                    break
            except GangaDiracError as err:
                logger.warning("Upload of input file as LFN %s to SE %s failed, trying another SE" % (new_lfn, SE)) 
    if not returnable:
        raise GangaException("Failed to upload input file to any SE")
    if should_del:
        os.unlink(os.path.join(localDir, namePattern))

    return returnable


def replicateJobFile(fileToReplicate):
    """
    A method to replicate a file to a random SE.
    """

    if not isinstance(fileToReplicate, DiracFile):
        raise GangaDiracError("Can only request replicas of DiracFiles. %s is not a DiracFile" % fileToReplicate)

    if len(fileToReplicate.locations)==0:
        fileToReplicate.getReplicas()

    trySEs = [SE for SE in getConfig('DIRAC')['allDiracSE'] if SE not in fileToReplicate.locations]
    random.shuffle(trySEs)
    success = None
    for SE in trySEs:
        if execute('checkSEStatus("%s", "%s")' % (SE, 'Write')):
            try:
                fileToReplicate.replicate(SE)
                success = True
                break
            except (GangaFileError, GangaDiracError) as err:
                logger.warning("Failed to replicate %s to %s. Trying another SE." % (fileToReplicate.lfn, SE))
    if not success:
        raise GangaException("Failed to replicate %s to any SE" % fileToReplicate.lfn)

def getInputFileDir(job):
    """
    Return the LFN remote dirname for this job
    """
    return os.path.join(DiracFile.diracLFNBase(job.backend.credential_requirements), 'GangaJob_%s/InputFiles' % job.fqid)


def check_creds(cred_req):
    """
    """
    try:
        credential_store[cred_req]
    except KeyError:
        credential_store.create(cred_req)


class GaudiExecDiracRTHandler(IRuntimeHandler):

    """The runtime handler to run plain executables on the Dirac backend"""

    data_file = 'data.py'

    def master_prepare(self, app, appmasterconfig):
        """
        Prepare the RTHandler for the master job so that applications to be submitted
        Args:
            app (GaudiExec): This application is only expected to handle GaudiExec Applications here
            appmasterconfig (unknown): Output passed from the application master configuration call
        """

        if app.autoDBtags and not app.getJobObject().inputdata[0].lfn.startswith('/lhcb/MC/'):
            logger.warning("This doesn't look like MC! Not automatically adding db tags.")
            app.autoDBtags = False

        cred_req = app.getJobObject().backend.credential_requirements
        check_creds(cred_req)

        inputsandbox, outputsandbox = master_sandbox_prepare(app, appmasterconfig)

        # If we are getting the metadata we need to make sure the summary.xml is added to the output sandbox if not there already.
        if app.getMetadata and not 'summary.xml' in outputsandbox:
            outputsandbox += ['summary.xml']

        # Check a previously uploaded input is there in case of a job copy
        if isinstance(app.uploadedInput, DiracFile):
            if app.uploadedInput.getReplicas() == {}:
                app.uploadedInput = None
                logger.info("Previously uploaded cmake target missing from Dirac. Uploading it again.")

        if not isinstance(app.uploadedInput, DiracFile):
            generateDiracInput(app)
            try:
                assert isinstance(app.uploadedInput, DiracFile)
            except AssertionError:
                raise ApplicationPrepareError("Failed to upload needed file, aborting submit. Tried to upload to: %s\nIf your Ganga installation is not at CERN your username may be trying to create a non-existent LFN. Try setting the 'DIRAC' configuration 'DiracLFNBase' to your grid user path.\n" % DiracFile.diracLFNBase(cred_req))
        
        rep_data = app.uploadedInput.getReplicas()
        try:
            assert rep_data != {}
        except AssertionError:
            raise ApplicationPrepareError("Failed to find a replica of uploaded file, aborting submit")


        if isinstance(app.jobScriptArchive, (DiracFile, LocalFile)):
            app.jobScriptArchive = None

        generateDiracScripts(app)

        try:
            assert isinstance(app.jobScriptArchive, DiracFile)
        except AssertionError:
            raise ApplicationPrepareError("Failed to upload needed file, aborting submit")
        rep_data = app.jobScriptArchive.getReplicas()
        try:
            assert rep_data != {}
        except AssertionError:
            raise ApplicationPrepareError("Failed to find a replica, aborting submit")

        #Check if the uploaded input already has replicas in case this is a copy of a job.
        if len(app.uploadedInput.locations)==0:
            app.uploadedInput.getReplicas()
        if len(app.uploadedInput.locations) >= 2:
            logger.debug("Uploaded input archive already at two locations, not replicating again")
            return
        else:
            replicateJobFile(app.uploadedInput)

        replicateJobFile(app.jobScriptArchive)

        return StandardJobConfig(inputbox=unique(inputsandbox), outputbox=unique(outputsandbox))


    def prepare(self, app, appsubconfig, appmasterconfig, jobmasterconfig):
        """
        Prepare the RTHandler in order to submit to the Dirac backend
        Args:
            app (GaudiExec): This application is only expected to handle GaudiExec Applications here
            appconfig (unknown): Output passed from the application configuration call
            appmasterconfig (unknown): Output passed from the application master_configure call
            jobmasterconfig (tuple): Output from the master job prepare step
        """
        cred_req = app.getJobObject().backend.credential_requirements
        check_creds(cred_req)

        # NB this needs to be removed safely
        # Get the inputdata and input/output sandbox in a sorted way
        inputsandbox, outputsandbox = sandbox_prepare(app, appsubconfig, appmasterconfig, jobmasterconfig)
        input_data,   parametricinput_data = dirac_inputdata(app)

        # We know we don't need this one
        inputsandbox = []

        job = app.getJobObject()

        # We can support inputfiles and opts_file here. Locally should be submitted once, remotely can be referenced.

        all_opts_files = app.getOptsFiles()

        for opts_file in all_opts_files:
            if isinstance(opts_file, DiracFile):
                inputsandbox += ['LFN:'+opts_file.lfn]
        # Sort out inputfiles we support
        for file_ in job.inputfiles:
            if isinstance(file_, DiracFile):
                inputsandbox += ['LFN:'+file_.lfn]
            elif isinstance(file_, LocalFile):
                if job.master is not None and file_ not in job.master.inputfiles:
                    shutil.copy(os.path.join(file_.localDir, file_.namePattern), app.getSharedPath())
                    inputsandbox += [os.path.join(app.getSharedPath(), file_.namePattern)]
            else:
                logger.error("Filetype: %s nor currently supported, please contact Ganga Devs if you require support for this with the DIRAC backend" % getName(file_))
                raise ApplicationConfigurationError("Unsupported filetype: %s with DIRAC backend" % getName(file_))

        master_job = job.master or job

        app.uploadedInput = master_job.application.uploadedInput
        app.jobScriptArchive = master_job.application.jobScriptArchive

        logger.debug("uploadedInput: %s" % app.uploadedInput)

        rep_data = app.uploadedInput.getReplicas()

        logger.debug("Replica info: %s" % rep_data)

        inputsandbox += ['LFN:'+app.uploadedInput.lfn]
        inputsandbox += ['LFN:'+app.jobScriptArchive.lfn]

        #Now add in the missing DiracFiles from the master job
        for file_ in master_job.inputfiles:
            if isinstance(file_, DiracFile) and 'LFN:'+file_.lfn not in inputsandbox:
                inputsandbox += ['LFN:'+file_.lfn]

        logger.debug("Input Sand: %s" % inputsandbox)

        logger.debug("input_data: %s" % input_data)

        outputfiles = [this_file for this_file in job.outputfiles if isinstance(this_file, DiracFile)]

        scriptToRun = getScriptName(app)
        # Already added to sandbox uploaded as LFN

        # This code deals with the outputfiles as outputsandbox and outputdata for us
        lhcbdirac_outputfiles = lhcbdirac_outputfile_jdl(outputfiles)

        # NOTE special case for replicas: replicate string must be empty for no
        # replication
        dirac_script = script_generator(lhcbdiracAPI_script_template(),
                                        DIRAC_IMPORT='from LHCbDIRAC.Interfaces.API.DiracLHCb import DiracLHCb',
                                        DIRAC_JOB_IMPORT='from LHCbDIRAC.Interfaces.API.LHCbJob import LHCbJob',
                                        DIRAC_OBJECT='DiracLHCb()',
                                        JOB_OBJECT='LHCbJob()',
                                        NAME=mangle_job_name(app),
                                        EXE=os.path.join('jobScript', scriptToRun),
                                        EXE_ARG_STR='',
                                        EXE_LOG_FILE='Ganga_GaudiExec.log',
                                        ENVIRONMENT=None,  # app.env,
                                        INPUTDATA=input_data,
                                        PARAMETRIC_INPUTDATA=parametricinput_data,
                                        OUTPUT_SANDBOX=API_nullifier(outputsandbox),
                                        OUTPUTFILESSCRIPT=lhcbdirac_outputfiles,
                                        OUTPUT_PATH="",  # job.fqid,
                                        OUTPUT_SE=[],
                                        PLATFORM=app.platform,
                                        SETTINGS=diracAPI_script_settings(app),
                                        DIRAC_OPTS=job.backend.diracOpts,
                                        REPLICATE='True' if getConfig('DIRAC')['ReplicateOutputData'] else '',
                                        # leave the sandbox for altering later as needs
                                        # to be done in backend.submit to combine master.
                                        # Note only using 2 #s as auto-remove 3
                                        INPUT_SANDBOX=repr([f for f in inputsandbox]),
                                        )

        # NB
        # inputsandbox here isn't used by the DIRAC backend as we explicitly define the INPUT_SANDBOX here!

        # Return the output needed for the backend to submit this job
        return StandardJobConfig(dirac_script, inputbox=[], outputbox=[])


#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#


def gaudiRun_script_template():
    """
    Script to return the contents of the command to be executed on the worker node
    """
    script_template = """#!/usr/bin/env python
'''Script to run Executable application'''
from __future__ import print_function
from os import listdir, environ, pathsep, getcwd, system
from mimetypes import guess_type
from contextlib import closing
import sys
import subprocess

def extractAllTarFiles(path):
    '''
    This extracts all extractable (g/b)zip files found in the top level dir of the WN
    '''
    for f in listdir(path):
        print("examining: %s" % f )
        file_type = guess_type(f)[1]
        if file_type in ['gzip', 'bzip2']:
            print("Extracting: %s" % f)
            if file_type == 'gzip':
                system("tar -zxf %s" % f)
            elif file_type == 'bzip2':
                system("tar -jxf %s" % f)

def pythonScript(scriptName):
    '''
    This allows us to run a custom python script on the WN
    '''
    script = '''
###WRAPPER_SCRIPT_WORKER_NODE###
'''
    return script

def flush_streams(pipe):
    '''
    This flushes the stdout/stderr streams to the stdout/stderr correctly
    '''
    with pipe.stdout or pipe.stderr:
        if pipe.stdout:
            for next_line in iter(pipe.stdout.readline, b''):
                print("%s" % next_line, file=sys.stdout, end='')
                sys.stdout.flush()
        if pipe.stderr:
            for next_line in iter(pipe.stderr.readline, b''):
                print("%s" % next_line, file=sys.stderr, end='')
                sys.stderr.flush()

# Main
if __name__ == '__main__':
    '''
    Main section of code for the GaudiExec script run on the WN
    '''
    # Opening pleasantries
    print("Hello from GaudiExec")
    print("Arrived at workernode: %s" % getcwd())
    print("#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/#")
    print("")

    print("CWD: %s" % getcwd())
    print("Files found on WN: %s" % (listdir('.')))

    if 'LHCb_release_area' not in environ:
        environ['LHCb_release_area'] = '/cvmfs/lhcb.cern.ch/lib/lhcb/'


    # Extract any/_all_ (b/g)zip files on the WN
    extractAllTarFiles('.')

    print("Executing: %s" % '###COMMAND###'+' '+' '.join(sys.argv[1:]))

    # Execute the actual command on the WN
    # NB os.system caused the entire stream to be captured before being streamed in some cases
    pipe = subprocess.Popen('###COMMAND###'+' '+' '.join(sys.argv[1:]), shell=True, stdin=subprocess.PIPE, stdout=subprocess.PIPE)

    # Flush the stdout/stderr as the process is running correctly
    flush_streams(pipe)

    # Wait for the process to finish executing
    pipe.wait()

    rc = pipe.returncode

    ###OUTPUTFILESINJECTEDCODE###

    # Final pleasantries
    print("")
    print("#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/#")
    print("Goodbye from GaudiExec")

    sys.exit(rc)
"""
    return script_template

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#

from GangaCore.GPIDev.Adapters.ApplicationRuntimeHandlers import allHandlers
allHandlers.add('GaudiExec', 'Dirac', GaudiExecDiracRTHandler)




#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#
import os
import os.path
import pprint
import tempfile
from GangaCore.Utility.Shell import Shell
from GangaCore.Utility.files import fullpath
from GangaCore.Utility.util import unique
import GangaCore.Utility.logging
import GangaCore.Utility.Config
from GangaCore.Core.exceptions import GangaException
import CMTUtils
import cmakeUtils
import subprocess
import time
import collections

logger = GangaCore.Utility.logging.getLogger()

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#

def make(self, arguments):

    if not self.newStyleApp is True:
        return CMTUtils.make(arguments)
    else:
        return cmakeUtils.make(arguments)

def get_user_platform(self, env=os.environ):

    if not self.newStyleApp is True:
        return CMTUtils.get_user_platform(env)
    else:
        return cmakeUtils.get_user_platform(env)


def update_project_path(self, user_release_area, env=os.environ):

    if not self.newStyleApp is True:
        return CMTUtils.update_project_path(user_release_area, env)
    else:
        return cmakeUtils.update_project_path(user_release_area, env)

def get_user_dlls(self, appname, version, user_release_area, platform, env):

    if not self.newStyleApp is True:
        return CMTUtils.get_user_dlls(appname, version, user_release_area, platform, env)
    else:
        return cmakeUtils.get_user_dlls(appname, version, user_release_area, platform, env)

def pyFileCollector(dir, file_list, subdir_dict, depth_cut, depth=0, zerodepth_pathlength=0):
    if zerodepth_pathlength == 0:
        zerodepth_pathlength = len(dir) + 1
    sub_pys = []
    for item in os.listdir(dir):
        file_path = os.path.join(dir, item)
        if (file_path.endswith('.py')):
            if os.path.exists(file_path):
                if depth == 0:
                    file_list.append(file_path)
                else:
                    sub_pys.append(file_path)
            else:
                logger.warning(
                    "File %s in %s does not exist. Skipping...", str(item), str(dir))
        elif os.path.isdir(file_path) and not os.path.islink(file_path):
            if depth >= depth_cut:
                continue
            pyFileCollector(
                file_path, file_list, subdir_dict, depth_cut, depth + 1, zerodepth_pathlength)

    if depth != 0:
        subdir_dict[dir[zerodepth_pathlength:]] = sub_pys


#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#

def shellEnv_cmd(cmd, environ=None, cwdir=None):
    pipe = subprocess.Popen(cmd,
                            shell=True,
                            env=environ,
                            cwd=cwdir,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)
    stdout, stderr = pipe.communicate()
    while pipe.poll() is None:
        time.sleep(0.5)
    return pipe.returncode, stdout, stderr


def shellEnvUpdate_cmd(cmd, environ=os.environ, cwdir=None):
    import tempfile
    import pickle
    f = tempfile.NamedTemporaryFile(mode='w+b')
    fname = f.name
    f.close()

    if not cmd.endswith(';'):
        cmd += ';'
    envdump = 'import os, pickle;'
    envdump += 'f=open(\'%s\',\'w+b\');' % fname
    envdump += 'pickle.dump(os.environ,f);'
    envdump += 'f.close();'
    envdumpcommand = 'python -c \"%s\"' % envdump
    cmd += envdumpcommand

    pipe = subprocess.Popen(cmd,
                            shell=True,
                            env=environ,
                            cwd=cwdir,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)
    stdout, stderr = pipe.communicate()
    while pipe.poll() is None:
        time.sleep(0.5)

    f = open(fname, 'r+b')
    environ = environ.update(pickle.load(f))
    f.close()
    os.system('rm -f %s' % fname)

    return pipe.returncode, stdout, stderr

    #\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/
    # NEW, better version?

CommandOutput = collections.namedtuple(
    'CommandOutput', ['returncode', 'stdout', 'stderr'])


class TimeoutException(Exception):

    def __init__(self, message):
        super(TimeoutException, self).__init__(message)


def run(cmd, env=None, cwd=None, timeout=None):
    proc = subprocess.Popen(cmd,
                            shell=True,
                            env=env,
                            cwd=cwd,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)
    if timeout is not None:
        time_start = time.time()
        while proc.poll() is None:
            if time.time() - time_start >= timeout:
                proc.kill()
                raise TimeoutException("Command '%s' timed out!" % cmd)
            time.sleep(0.5)
    stdout, stderr = proc.communicate()
    return CommandOutput(proc.returncode, stdout, stderr)


def runUpdate(cmd, env=None, cwd=None, timeout=None):
    (fd, filename) = tempfile.mkstemp()
    command_output = run(cmd + '; printenv &> ' + filename, env, cwd, timeout)
    with os.fdopen(fd, 'r') as file:
        env.update(dict([tuple(line.split('=', 1))
                         for line in file.read().splitlines()]))
    return command_output


#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#
def fillPackedSandbox(sandbox_files, destination):
    """Put all sandbox_files into tarball called name and write it into to the destination.
    Arguments:
    'sandbox_files': a list of File or FileBuffer objects.
    'destination': a string representing the destination filename
    Return: a list containing a path to the tarball
    """
    if not sandbox_files:
        return []

    # Generalised version from Ganga/Core/Sandbox/Sandbox.py

    import tarfile
    import stat

    #tf = tarfile.open(destination,"w:gz")

    # "a" = append with no compression
    # creates file if doesn't exist
    # cant append to a compressed tar archive so must compress later
    dir, filename = os.path.split(destination)
    if not os.path.isdir(dir):
        os.makedirs(dir)
    tf = tarfile.open(destination, "a")
    tf.dereference = True  # --not needed in Windows

    for f in sandbox_files:
        try:
            contents = f.getContents()   # is it FileBuffer?

        except AttributeError:         # File
            try:
                fileobj = file(f.name)
            except:
                raise GangaException("File %s does not exist." % f.name)
            tinfo = tf.gettarinfo(
                f.name, os.path.join(f.subdir, os.path.basename(f.name)))

        else:                          # FileBuffer
            from StringIO import StringIO
            fileobj = StringIO(contents)

            tinfo = tarfile.TarInfo()
            tinfo.name = os.path.join(f.subdir, os.path.basename(f.name))
            import time
            tinfo.mtime = time.time()
            tinfo.size = fileobj.len

        if f.isExecutable():
            tinfo.mode = tinfo.mode | stat.S_IXUSR
        tf.addfile(tinfo, fileobj)

    tf.close()

    return [destination]


def gzipFile(filename, outputfilename=None, removeOriginal=False):
    """
    This method creates a new compressed version of the inputfile compressed with gzip
    Args:
        filename (str): This is the file which is being compressed
        outputfilename (str): This is the (optional) name of the outputfile object which is created
        removeOriginal (bool): Should the original be destroyed after the compressed version is made?
    """
    import gzip
    if not outputfilename:
        outputfilename = filename + '.gz'
    f_in = open(filename, 'rb')
    f_out = gzip.open(outputfilename, 'wb')
    f_out.writelines(f_in)
    f_out.close()
    f_in.close()
    if removeOriginal:
        os.system('rm -f %s' % filename)




######################################################
# RATProd.py
# ---------
# Author: Matt Mottram
#         <m.mottram@qmul.ac.uk>
#
# Description:
#    Ganga application for SNO+ production.
#
# Runs tagged RAT releases on a given backend via ratProdRunner.py
#
# Classes:
#  - RATProd: production applicaiton
#  - RATSplitter: splitter to create subjobs for subruns
#
# Revision History:
#  - 03/11/12: M. Mottram: first revision with proper documentation!
#  - 26/08/14: M. Mottram: moved RTHandlers to RTRATProd.py
# 
######################################################

import os

from GangaCore.Core.exceptions import ApplicationConfigurationError

from GangaCore.GPIDev.Adapters.IApplication import IApplication
from GangaCore.GPIDev.Adapters.ISplitter import ISplitter
from GangaCore.GPIDev.Schema import *

from GangaCore.GPIDev.Lib.File import *

###################################################################

config = GangaCore.Utility.Config.makeConfig('defaults_RATProd','Defaults for the RATProd application')

config.addOption('local_softwareEnvironment', None, 'Local snoing-install directory (or directory with env_rat-x.y.sh files)')
config.addOption('local_environment', [], 'Environment options required to run on local or batch system')
config.addOption('local_outputDir', None, '*Default* output directory if running on a batch or local system (can override)')
config.addOption('grid_outputDir', None, '*Defult* output directory if running on system with grid storage (can override)')

# Assume that the applications should come from the same GangaSNOplus directory        
_app_directory = os.path.dirname(__file__)

###################################################################

class RATProd(IApplication):
    """The RAT job handler for data production and processing"""

    #_schema is required for any Ganga plugin
    _schema = Schema(Version(1,1), {            
            'environment'       : SimpleItem(defvalue=None,doc='list of strings with the commands to setup the correct backend environment, or single string location of a file with the appropriate commands (if necessary)',typelist=['list','str','type(None)']),
            'discardOutput'     : SimpleItem(defvalue=False,doc='Do not store the output: default False',typelist=['bool']),
            'inputDir'          : SimpleItem(defvalue='',doc='Provide a relative path (to the base dir - dependent on the backend) for any inputs',
                                             typelist=['str']),
            'inputFiles'        : SimpleItem(defvalue=[],doc='Input file names (must be a list)',
                                             typelist=['list','str']),#don't use if splitting
            'outputLog'         : SimpleItem(defvalue='rat_output.log',doc='Log file name (only used if RAT is run)',
                                             typelist=['str']),
            'outputDir'        : SimpleItem(defvalue='',doc='Provide a relative path (to the base dir - dependent on the backend) for the file to be archived',
                                             typelist=['str']),
            'outputFiles'       : SimpleItem(defvalue=[],doc='Output file names (can be a list)',
                                             typelist=['list','str']),#don't use if splitting
            'prodScript'        : SimpleItem(defvalue='',doc='String pointing to the script file to run.  Can be set by splitter.',
                                             typelist=['str']),
            'rat_db_name'       : SimpleItem(defvalue=None, doc='RAT db name', typelist=['str', 'type(None)']),
            'rat_db_pswd'       : SimpleItem(defvalue=None, doc='RAT db password', typelist=['str', 'type(None)']),
            'rat_db_protocol'   : SimpleItem(defvalue=None, doc='RAT db protocol', typelist=['str', 'type(None)']),
            'rat_db_url'        : SimpleItem(defvalue=None, doc='RAT db password', typelist=['str', 'type(None)']),
            'rat_db_user'       : SimpleItem(defvalue=None, doc='RAT db password', typelist=['str', 'type(None)']),
            'softwareEnvironment': SimpleItem(defvalue=None,doc='Software environment file, required if running on a non-LCG backend',
                                             typelist=['str', 'type(None)']),
            'ratMacro'          : SimpleItem(defvalue='',doc='String pointing to the macro file to run.  Can be set by splitter.',
                                             typelist=['str']),#shouldn't this be in the input sandbox?
            'ratVersion'        : SimpleItem(defvalue='4',doc='RAT version tag, necessary to setup environment (even if not running RAT)',
                                             typelist=['str']),
            'useDB'             : SimpleItem(defvalue=False,doc='Use the RAT database (snopl.us)?',typelist=['bool']),
            })
    
    _category = 'applications'
    _name = 'RATProd'

    def configure(self,masterappconfig):
        '''Configure method, called once per job.
        '''
        logger.debug('RAT::RATProd configure ...')

        job = self._getParent()
        masterjob = job._getParent()

        if self.prodScript=='' and self.ratMacro=='':
            logger.error('prodScript or ratMacro not defined')
            raise Exception
        elif self.prodScript!='' and self.ratMacro!='':
            logger.error('both prodScript and ratMacro are defined')
            raise Exception
        if self.useDB:
            if not config['rat_db_pswd']:
                logger.error('Need a password in order to contact the ratdb database')
                raise Exception

        #The production script is added in, line by line, into the submission script
        #job.inputsandbox.append(File(self.prodScript))

        if self.ratMacro!='':
            #decimated=self.ratMacro.split('/')
            #macFile=decimated[len(decimated)-1]#for when I thought we needed args to go with the script
            job.inputsandbox.append(File(self.ratMacro))
            #Always run rat with a log called rat.log
            job.outputsandbox.append('rat.log')
        else:
            job.inputsandbox.append(File(self.prodScript))
        job.outputsandbox.append('return_card.js')

        #we want a list of files - if we only have one (i.e. a string) then just force into a list
        if type(self.inputFiles) is str:
            self.inputFiles = [self.inputFiles]
        if type(self.outputFiles) is str:
            self.outputFiles = [self.outputFiles]

        return(None,None)
                     
###################################################################

class RATSplitter(ISplitter):
    '''Splitter for RAT jobs.  Essentially an ArgSplitter, but specifically for the output files.
    '''
    _name = "RATSplitter"
    _schema = Schema(Version(1,0), {
            'outputFiles' : SimpleItem(defvalue=[],typelist=['list','str'],sequence=1,doc='A list of lists for specifying output files.  Only works with the RATProd application.'),
            'inputFiles'  : SimpleItem(defvalue=[],typelist=['list','str'],sequence=1,doc='A list of lists for specifying intput files.  Only works with the RATProd application.'),
            'prodScript'  : SimpleItem(defvalue=[],typelist=['str'],sequence=1,doc='A list specifying the scripts (if used).'),
            'ratMacro'    : SimpleItem(defvalue=[],typelist=['str'],sequence=1,doc='A list specifying the macros (if used)'),
            } )
    
    def split(self,job):
        
        subjobs = []

        fin=True
        fout=True
        nFiles=0

        #check if there are input or output files
        if len(self.inputFiles)==0:
            fin=False
        else:
            nFiles = len(self.inputFiles)
        if len(self.outputFiles)==0:
            fout=False
        else:
            nFiles = len(self.outputFiles)
        nScript = len(self.prodScript)
        nMacro = len(self.ratMacro)
        if nScript!=0 and nMacro!=0:
            logger.error('cannot provide both scripts and macros, must be either or')
            raise Exception
        if nScript==0 and nMacro==0:
            logger.error('must provide EITHER macros or scripts')
            raise Exception

        #if neither, what the hell are we splitting for!?
        if fin==False and fout==False:
            logger.error('Why are you splitting?')
            raise Exception

        #if we have both input and output, they should have the same # entries
        if fin==True and fout==True:
            if len(self.inputFiles)!=len(self.outputFiles):
                logger.error('input/output numbers dont match')
                raise Exception
        if nScript!=0:
            if nScript!=nFiles:
                logger.error('script/file numbers dont match')
                raise Exception
        if nMacro!=0:
            if nMacro!=nFiles:
                logger.error('macro/file numbers dont match')
                raise Exception

        for i in range(nFiles):
            #nFiles should have the same number as either (or both) #input or #output
            j = self.createSubjob(job)
            # Add new arguments to subjob
            if fout is True:
                #if type(self.outputFiles[i]) is not list:
                #    logger.error('need a list of files! %s' % type(self.outputFiles[i]))
                j.application.outputFiles=self.outputFiles[i]
            if fin is True:
                #if type(self.inputFiles[i]) is not list:
                #    logger.error('need a list of files! %s' % type(self.inputFiles[i]))
                j.application.inputFiles=self.inputFiles[i]
            if nMacro!=0:
                j.application.ratMacro=self.ratMacro[i]
            else:
                j.application.prodScript=self.prodScript[i]
            subjobs.append(j)
        return subjobs
    
###################################################################

from GangaCore.GPIDev.Adapters.ApplicationRuntimeHandlers import allHandlers
from GangaSNOplus.Lib.RTHandlers.RTRATProd import RTHandler, WGRTHandler, LCGRTHandler

allHandlers.add('RATProd','Local', RTHandler)
allHandlers.add('RATProd','PBS', RTHandler)
allHandlers.add('RATProd','SGE', RTHandler)
allHandlers.add('RATProd','Condor', RTHandler)
allHandlers.add('RATProd','LCG', LCGRTHandler)
allHandlers.add('RATProd','TestSubmitter', RTHandler)
allHandlers.add('RATProd','Interactive', RTHandler)
allHandlers.add('RATProd','Batch', RTHandler)
allHandlers.add('RATProd','WestGrid', WGRTHandler)

logger = GangaCore.Utility.logging.getLogger()



################################################################################
# Ganga - a computational task management tool for easy access to Grid resources
# http://cern.ch/ganga
#
# $Id: driver.py,v 1.3 2008-12-04 13:17:58 moscicki Exp $
#
# Copyright (C) 2003-2007 The Ganga Project
#
# This file is part of GangaCore. 
#
# Ganga is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Ganga is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
################################################################################

"""
Test driver used to executed GPI, GPIM and PyUnit tests 
This file is executed as a Ganga script so we can access to all GPI objects
"""

import os,os.path,sys,new

from GangaCore.Utility.Config import getConfig
myconfig = getConfig('TestingFramework')
from GangaCore.Utility.logging import getLogger
#logger
logger=getLogger('GangaTest.Framework')

class GPIRunner:
        """
        load/setup/run/tearDown encapsulation for a GPI test
        Note: These operations are executed in the same ganga process
        """
        def __init__(self, testPath, testName):
                self.testPath = testPath
                self.testName = testName
        
        def load(self):
                """
                        NOP: the testPath is sufficient for this type of tests
                """
                return True
        
        def setup(self):
                """
                Apart from .startup script which is executed in a different Ganga session
                optionally the job registry can be cleaned up.
                """
                #auto cleanup of job registry as a pre-step of the initialization (optionally)
                cleanup()
        
        def run(self): 
                """
                Execute file in a Ganga session
                """
                import os.path
                testFile = os.path.split(testPath)[1]           
                execfile( testFile , GangaCore.Runtime._prog.local_ns )
                
        def tearDown(self):
                """
                RUNNING CLEANUP IN THE NEXT SESSION CAUSES MONITORING ERRORS!!!!!
                let's NEVER, NEVER, EVER! rely on this - rcurrie
                """
                cleanup()
        
                
class UnitRunner:
        """
        load/setup/run/tearDown pipe-line for a GPIM/PY tests
        """
        
        def __init__(self, testPath, testName):
                self.testPath = testPath
                self.testName = testName
        
        def load(self):
                self.__instance = None
                try:                    
                        exec_module_code(self.testPath,GangaCore.Runtime._prog.local_ns)
                        testClass=os.path.splitext(os.path.basename(self.testPath))[0]
                        clazz = GangaCore.Runtime._prog.local_ns[testClass]# getattr(GangaCore.Runtime._prog.local_ns.__dict__, testClass)                  
                        self.__instance = clazz()
                except Exception as e:
                        logger.error("Cannot load test")
                        import traceback
                        traceback.print_exc(file=sys.stderr)
                        self.__instance=None
                        return False
                return True
        
        def setup(self):
                if hasattr(self.__instance,"setUp") and "setUp" in sys.argv:
                        #auto cleanup of job registry as a pre-step of the initialization
                        cleanup()
                        getattr(self.__instance,"setUp")()
                        
        def run(self):
                getattr(self.__instance,self.testName)()

        def tearDown(self):
                try:
                        if hasattr(self.__instance,"tearDown") and "tearDown" in sys.argv:
                                getattr(self.__instance,"tearDown")()
                except Exception as e:
                        logger.error("Cannot tear down gracefully the test (%s)" % str(e))
                        import traceback
                        traceback.print_exc()

# Use the modules from testoob for GPIPRunner.
sys.path.append(os.path.join(os.getenv('PYTF_TOP_DIR','').split(':')[0],'pytf'))
from pytf.testoob.extracting import full_extractor as _full_extractor
from pytf.testoob.running.convenience import _create_reporter_proxy
from pytf.testoob.running.simplerunner import SimpleRunner
from pytf.testoob.running.convenience import apply_decorators
from pytf.testoob.reporting.colored import ColoredTextReporter
from pytf.testoob.reporting import XMLFileReporter

import unittest
from GangaTest.Framework.tests import GPIPPreparationTestCase, SimpleRunnerControl, FailedTestsException

class GPIPRunner:
        def __init__(self, testPath, outputPath,timeout,description,releaseTest,report_outputpath,parent_report_name):
                self.testPath = testPath
                #self.testName = testName
                self.outputPath = outputPath
                self.timeout = timeout
                self.description = description # description template - 'GangaDummy/old_test/Root/TestRoot/%s [GPIP]'
                self.releaseTest = releaseTest # is release test mode enable 
                self.report_outputpath = report_outputpath # release report directory 
                self.parent_report_name = parent_report_name # indicate the parent report name(suite name)
                self.pytest_name = os.path.basename(self.outputPath.split('%s')[0].strip('.'))
                self.gpiptest_prefix = self.pytest_name.replace('.', '/')
                self._testinstances = []
                self._xml_report_ext = '%s.xml_merged'

        def load(self):
                self.__instance = None
                try:
                        exec_module_code(self.testPath,GangaCore.Runtime._prog.local_ns)
                        testClass=os.path.splitext(os.path.basename(self.testPath))[0]
                        clazz = GangaCore.Runtime._prog.local_ns[testClass]# getattr(GangaCore.Runtime._prog.local_ns.__dict__, testClass)
                        self.clazz = clazz
                        self.__instance = clazz()
                        self.testsuite = self._getTestSuite()
                except Exception as e:
                        logger.error("Cannot load test")
                        import traceback
                        traceback.print_exc(file=sys.stderr)
                        self.__instance=None
                        return False
                return True

        def setup(self):
                if hasattr(self.__instance,"setUp") and "setUp" in sys.argv:
                        #auto cleanup of job registry as a pre-step of the initialization
                        cleanup()
                        for instance in self._testinstances:
                            getattr(instance,"setUp")()

        def run(self):
                """
                Execute file in a Ganga session
                """
                import time

                from pytf.testoob.running.fixture_decorators import BaseFixture
                fixture_decorators = [BaseFixture]

                test_extractor = apply_decorators(_full_extractor, [])

                runners_control = []
                
                failedTests = []

                start = time.time()

                # self.testsuite must be initialized in load()
                for fixture in test_extractor(self.testsuite):
                        testName = fixture.method_name
                        decorated_fixture = apply_decorators(fixture, fixture_decorators)
                        runner = self._generate_runner(testName)
                        runner_control = SimpleRunnerControl(testName, runner, decorated_fixture, fixture_decorators, self.timeout)
                        if runner_control is not None: runners_control.append(runner_control)

                for run_control in runners_control:
                    run_status = run_control.runPreparationTestCase()
                    if not run_status:
                        run_control.setFinished(True) # Fail occurred while preparing the job, stop the test and write out the report.
                        if run_control.preparationError:
                            failedTests.append(["%s:%s" % (self.gpiptest_prefix, run_control.testName), run_control.preparationError])
                        else:
                            failedTests.append(["%s:%s" % (self.gpiptest_prefix, run_control.testName), 'Failed in preparation somehow.'])

                all_done = False

                #start = time.time()

                duration = start - time.time()

                while not all_done and not (duration > self.timeout):
                    all_done = True
                    for run_control in runners_control:
                        if not run_control.isReadyForCheck() and not run_control.isFinished():
                            all_done = False
                    if not all_done:
                        end = time.time()
                        duration = end - start
                    time.sleep(1)

                if duration > self.timeout:
                    for run_control in runners_control:
                        if not run_control.isReadyForCheck():
                            run_control.setTimeout(True) # Mark the jobs which are still running with timeout 'True' flag.

                for run_control in runners_control:
                    if not run_control.isFinished():
                        if not run_control.runCheckTest():
                            if run_control.error:
                                failedTests.append(["%s:%s" % (self.gpiptest_prefix, run_control.testName), run_control.errorTraceback])
                            elif run_control.timeoutError:
                                failedTests.append(["%s:%s" % (self.gpiptest_prefix, run_control.testName), run_control.errorTraceback])
                            elif run_control.runCheckError:
                                failedTests.append(["%s:%s" % (self.gpiptest_prefix, run_control.testName), run_control.errorTraceback])
                            else:
                                failedTests.append(["%s:%s" % (self.gpiptest_prefix, run_control.testName), 'Failed in running check test somehow.'])

                logger.info("[%d failed tests]" % len(failedTests))
                if len(failedTests) > 0:
                    failedTestsException = FailedTestsException(failedTests)
                    import traceback
                    traceback.print_exc()
                    raise failedTestsException


        def _generate_runner(self, testName):
            from pytf.testoob.running.simplerunner import SimpleRunner
            runner = SimpleRunner()

            verbosity = 1
            immediate = False
            coverage = (None, None)
            
            reporters = []

            if self.releaseTest:
                runner.xml_filename = self._xml_report_ext % os.path.join(self.report_outputpath, '%s_%s_%s' % (self.parent_report_name,self.pytest_name,testName))
                xml_reporter = XMLFileReporter(filename=runner.xml_filename)
                reporters.append(xml_reporter)
            else:
                runner.xml_filename = None

            
            coloredReporter = ColoredTextReporter(
                              verbosity=verbosity,
                              immediate=immediate,
                              descriptions=1,
                              stream=sys.stderr)

            reporters.append(coloredReporter)
            

            runner.reporter = _create_reporter_proxy(reporters, None, threads=True)

            return runner


        def _getTestSuite(self):
                attrs = dir(self.clazz)
                testcase = None
                testcases = []
                for test in attrs:
                        stest = str(test)
                        if stest.startswith('test'):
                                testcase = GPIPPreparationTestCase(self._get_tester_instance(), self.description % stest, stest, self.timeout, self.outputPath % stest)
                                if testcase is not None: testcases.append(testcase)
                return unittest.TestSuite(testcases)

        def _get_tester_instance(self):
            instance = self.clazz()
            self._testinstances.append(instance)
            return instance

        def tearDown(self):
                try:
                        if hasattr(self.__instance,"tearDown") and "tearDown" in sys.argv:
                            for instance in self._testinstances:
                                getattr(instance,"tearDown")()
                except Exception as e:
                        logger.error("Cannot tear down gracefully the test (%s)" % str(e))
                        import traceback
                        traceback.print_exc()

#utils 
def cleanup():   
    if myconfig['AutoCleanup'] == True:
        logger.info("[TestingFramework]AutoCleanup=True")
        logger.info("cleaning up job repository...")
        # explicit removal of jobs is not required but makes one test more
        try:
            for id in jobs.ids():
                jobs(id).remove(force=True)
            jobs.remove(force=True)

            for id in templates.ids():
                templates(id).remove(force=True)
            templates.remove(force=True)
            logger.info("regular cleanup done.")
        except Exception as e:
            logger.exception("Cleanup failed (%s)... will delete repository and workspace dirs manually "%str(e))

        logger.info('performing "hard" cleanup')

        if hasattr(jobs,'clean'):
                jobs.clean(confirm=True, force=True)
        if hasattr(templates,'clean'):
                templates.clean(confirm=True, force=True)

        try:
            lock_files = GangaCore.Core.GangaRepository.SessionLock.getGlobalSessionFiles()

            for i in lock_files:
                if i.endswith('global_lock'):
                    import os.path
                    lockfile_path = os.path.dirname( i )
                    break

            sessions = os.listdir(lockfile_path)

            for i in sessions:
                j = os.path.join( lockfile_path, i )
                if j not in lock_files:
                    try:
                        os.close( j )
                    except:
                        pass
                    try:
                        os.unlink( j )
                    except:
                        pass

        except:
            pass


def read_file(filename):
        f = open(filename)
        try: return "%s\n"%f.read()
        finally: f.close()

def exec_module_code(path,ns=None ):
        """
            loads the code from path and execute it in the given namespace
            If ns is None the code is executed in the newly created module namespace
        """
        (testdir,filename) = os.path.split(path)    
        cwd = os.getcwd()
        if testdir:
                try:
                        os.chdir(testdir)
                except:
                        pass

        name     = os.path.splitext(filename)[0]
        code_str = read_file(filename)
        code     = compile(code_str,filename,'exec')
        (name,ext) = os.path.splitext(filename)
        if name in sys.modules:
                mod = sys.modules[name] # necessary for reload()
        else:
                mod = new.module(name)
                sys.modules[name] = mod
        mod.__name__ = name
        mod.__file__ = filename
        if ns is None:
                exec(code, mod.__dict__)
        else:
                exec(code, ns)
        
        #restore cwd
        os.chdir(cwd)       
        return mod

def parse_args(sys_argv):    
        opts = {}
        params = []    
        for arg in sys_argv:
                if arg.startswith("--"):            
                        list = arg.split('=')
                        opt = list[0]
                        try:
                                val = list[1]
                        except IndexError:
                                val = None                          
                        opts[opt]=val
                else:
                        params.append(arg)
        return params,opts


if __name__=="__main__":
        
        if len(sys.argv)<4:
                logger.error("Invalid usage of program. At least three parameters are required")
                sys.exit(1)     
        
        param,opt = parse_args(sys.argv[1:])    
        testPath=param[0]
        try:
                testName=param[1]
        except IndexError:
                testName = None
        coverage_report = opt.get("--coverage-report", None) 
        #test type: GPI, GPIM or PY    
        test_type = opt.get("--test-type", "gpi")
        if test_type == 'gpi':
                testRunner = GPIRunner(testPath,testName)
        elif test_type == 'gpip':
                #testRunner = UnitRunner(testPath,testName)
                testRunner = GPIPRunner(testPath,opt.get('--output-path',None),int(opt.get('--timeout',120)),opt.get('--description'),opt.get('--release-test',False),opt.get('--report-outputpath',None),opt.get('--parent-report-name',None))
        elif test_type in ['py','gpim']:
                testRunner = UnitRunner(testPath,testName) 
        
        #1. LOAD
        if not testRunner.load():
                sys.exit(1)
        success = True
        try:
                #2. SETUP
                testRunner.setup()      
                if coverage_report and coverage_report!='None':
                        import figleaf
                        figleaf.start()
                #3. RUN
                testRunner.run()

                if coverage_report and coverage_report!='None':
                        try:
                                figleaf.stop()
                                figleaf.write_coverage(coverage_report)
                        except NameError:
                                pass
                #4. TEAR-DOWN
                testRunner.tearDown()   
        except Exception as e:
                import traceback
                traceback.print_exc(file=sys.stderr)
                success = False
                if coverage_report and coverage_report!='None':
                        try:
                                figleaf.stop()
                                figleaf.write_coverage(coverage_report)
                        except:
                                pass

        try:
                ## Disable internal services such as monitoring and other tasks
                from GangaCore.Core.InternalServices import Coordinator
        #        if Coordinator.servicesEnabled:
        #                Coordinator.disableInternalServices()

                from GangaCore.Core.InternalServices import ShutdownManager
                ShutdownManager._ganga_run_exitfuncs()

        except:
                pass

        #sys.exit(not success)
        import os
        os._exit(not success)

        # If this crops up again that this stalls on exit the ONLY reliable soluition
        # I found was to call os._exit( not sucess ) which is horrible and potentially dangerous

#$Log: not supported by cvs2svn $
#Revision 1.2  2008/11/26 08:30:35  moscicki
#GPIP (parallel) tests from Mason
#untabified the driver.py file
#
#Revision 1.1  2008/07/17 16:41:35  moscicki
#migration of 5.0.2 to HEAD
#
#the doc and release/tools have been taken from HEAD
#
#Revision 1.1.2.7  2008/04/18 13:50:22  moscicki
#resetAll() method wipes all the repository
#
#Revision 1.1.2.6  2008/04/04 16:03:23  moscicki
#fixed import in cleanup (Will)
#
#Revision 1.1.2.5  2008/03/13 15:53:32  amuraru
#forcibly remove repos and wspace dir in cleanup code
#
#Revision 1.1.2.4  2008/02/29 15:33:40  amuraru
#JobRegistry fix
#
#Revision 1.1.2.3  2007/12/19 17:37:03  amuraru
#removed the cleanup file and moved the cleanup code in test driver
#fixed the namespace problems when executing test-caseswq
#



"""
Internal initialization of the repositories.
"""

from pipes import quote
import os.path
from GangaCore.Utility.Config import getConfig, setConfigOption
from GangaCore.Utility.logging import getLogger
from GangaCore.Utility.files import expandfilename, fullpath
from GangaCore.Core.GangaRepository import getRegistries
from GangaCore.Core.GangaRepository import getRegistry
from GangaCore.Core.exceptions import GangaException

config = getConfig('Configuration')
logger = getLogger()

_runtime_interface = None

def requiresAfsToken():
    # Were we executed from within an AFS folder
    return fullpath(getLocalRoot(), True).find('/afs') == 0


def getLocalRoot():
    # Get the local top level directory for the Repo
    if config['repositorytype'] in ['LocalXML', 'LocalAMGA', 'LocalPickle', 'SQLite']:
        return os.path.join(expandfilename(config['gangadir'], True), 'repository', config['user'], config['repositorytype'])
    else:
        return ''

def getLocalWorkspace():
    # Get the local top level dirtectory for the Workspace
    if config['repositorytype'] in ['LocalXML', 'LocalAMGA', 'LocalPickle', 'SQLite']:
        return os.path.join(expandfilename(config['gangadir'], True), 'workspace', config['user'], config['repositorytype'])
    else:
        return ''


started_registries = []

partition_warning = 95
partition_critical = 99

def checkDiskQuota():
    # Throw an error atthe user if their AFS area is (extremely close to) full to avoid repo corruption
    import subprocess

    repo_partition = getLocalRoot()
    repo_partition = os.path.realpath(repo_partition)

    work_partition = getLocalWorkspace()
    work_partition = os.path.realpath(work_partition)

    folders_to_check = [repo_partition, work_partition]

    home_dir = os.environ['HOME']
    to_remove = []
    for partition in folders_to_check:
        if not os.path.exists(partition):
            if home_dir not in folders_to_check:
                folders_to_check.append(home_dir)
            to_remove.append(partition)

    for folder in to_remove:
        folders_to_check.remove(folder)

    for data_partition in folders_to_check:

        if fullpath(data_partition, True).find('/afs') == 0:
            quota = subprocess.Popen(" ".join(['fs', 'quota', '%s' % quote(data_partition)]), shell=True, stdout=subprocess.PIPE)
            output = quota.communicate()[0]
            logger.debug("fs quota %s:\t%s" % (quote(data_partition), output))
        else:
            df = subprocess.Popen(["df", '-Pk', '%s' % data_partition], stdout=subprocess.PIPE)
            output = df.communicate()[0]

        try:
            global partition_warning
            global partition_critical
            quota_percent = output.split('%')[0]
            if int(quota_percent) >= partition_warning:
                logger.warning("WARNING: You're running low on disk space, Ganga may stall on launch or fail to download job output")
                logger.warning("WARNING: Please free some disk space on: %s" % data_partition)
            if int(quota_percent) >= partition_critical and config['force_start'] is False:
                logger.error("You are crtitically low on disk space!")
                logger.error("To prevent repository corruption and data loss we won't start GangaCore.")
                logger.error("Either set your config variable 'force_start' in .gangarc to enable starting and ignore this check.")
                logger.error("Or, make sure you have more than %s percent free disk space on: %s" %(100-partition_critical, data_partition))
                raise GangaException("Not Enough Disk Space!!!")
        except GangaException as err:
            raise
        except Exception as err:
            logger.debug("Error checking disk partition: %s" % err)

    return

def bootstrap_getreg():
    # Get the list of registries sorted in the bootstrap way
    
    # ALEX added this as need to ensure that prep registry is started up BEFORE job or template
    # or even named templated registries as the _auto__init from job will require the prep registry to
    # already be ready. This showed up when adding the named templates.
    def prep_filter(x, y):
        if x.name == 'prep':
            return -1
        return 1

    return [registry for registry in sorted(getRegistries(), prep_filter)]

def bootstrap_reg_names():
    # Get the list of registry names
    all_reg = bootstrap_getreg()
    return [reg.name for reg in all_reg]

def bootstrap():
    # Bootstrap for startup and setting of parameters for the Registries
    retval = []

    try:
        checkDiskQuota()
    except GangaException as err:
        raise
    except Exception as err:
        logger.error("Disk quota check failed due to: %s" % err)

    for registry in bootstrap_getreg():
        if registry.name in started_registries:
            continue
        if not hasattr(registry, 'type'):
            registry.type = config["repositorytype"]
        if not hasattr(registry, 'location'):
            registry.location = getLocalRoot()
        logger.debug("Registry: %s" % registry.name)
        logger.debug("Loc: %s" % registry.location)
        registry.startup()
        logger.debug("started " + registry.info(full=False))

        started_registries.append(registry.name)
        proxied_registry_slice = registry.getProxy()
        retval.append((registry.name, proxied_registry_slice, registry.doc))

    # Assuming all registries are started for all instances of Ganga atm
    # Avoid ever, ever, calling the repository from behind the registry. This allows for all forms of bad behaviour
    # This is a form of trade off but still is something which should be strongly discouraged!
    # TODO investigate putting access to the repository behind a getter/setter which keeps the registry locked
    other_sessions = bootstrap_getreg()[0].repository.get_other_sessions()
    if other_sessions:
        # Just print this from 1 repo only so chose the zeorth, nothing special
        logger.warning("%i other concurrent sessions:\n * %s" % (len(other_sessions), "\n * ".join(other_sessions)))
        logger.warning("Multiple Ganga sessions detected. The Monitoring Thread is being disabled.")
        logger.warning("Type 'enableMonitoring' to restart")
        setConfigOption('PollThread', 'autostart', False)

    return retval


def updateLocksNow():
    # Update all of the file locks for the registries

    logger.debug("Updating timestamp of Lock files")
    for registry in getRegistries():
        registry.updateLocksNow()
    return


def shutdown():
    # Shutdown method for all repgistries in order
    from GangaCore.Utility.logging import getLogger
    logger = getLogger()
    logger.info('Registry Shutdown')
    #import traceback
    #traceback.print_stack()

    # Flush all repos before we shut them down
    flush_all()

    # shutting down the prep registry (i.e. shareref table) first is necessary to allow the closedown()
    # method to perform actions on the box and/or job registries.
    logger.debug(started_registries)

    all_registries = getRegistries()

    try:
        if 'prep' in started_registries:
            registry = getRegistry('prep')
            registry.shutdown()
            # in case this is called repeatedly, only call shutdown once
            started_registries.remove(registry.name)
    except Exception as err:
        logger.error("Err: %s" % err)
        logger.error("Failed to Shutdown prep Repository!!! please check for stale lock files")
        logger.error("Trying to shutdown cleanly regardless")

    for registry in getRegistries():
        thisName = registry.name
        try:
            if not thisName in started_registries:
                continue
            # in case this is called repeatedly, only call shutdown once
            started_registries.remove(thisName)
            registry.shutdown()  # flush and release locks
        except Exception as x:
            logger.error("Failed to Shutdown Repository: %s !!! please check for stale lock files" % thisName)
            logger.error("%s" % x)
            logger.error("Trying to Shutdown cleanly regardless")


    for registry in all_registries:

        my_reg = [registry]
        if hasattr(registry, 'metadata'):
            if registry.metadata:
                my_reg.append(registry.metadata)

        assigned_attrs = ['location', 'type']
        for this_reg in my_reg:
            for attr in assigned_attrs:
                if hasattr(registry, attr):
                    delattr(registry, attr)

    from GangaCore.Core.GangaRepository.SessionLock import removeGlobalSessionFiles, removeGlobalSessionFileHandlers
    removeGlobalSessionFileHandlers()
    removeGlobalSessionFiles()

    removeRegistries()

def flush_all():
    # Flush all registries in their current state with all dirty knowledge going to disk
    from GangaCore.Utility.logging import getLogger
    logger = getLogger()
    logger.debug("Flushing All repositories")

    for registry in getRegistries():
        thisName = registry.name
        try:
            if registry.hasStarted() is True:
                logger.debug("Flushing: %s" % thisName)
                registry.flush_all()
        except Exception as err:
            logger.debug("Failed to flush: %s" % thisName)
            logger.debug("Err: %s" % err)


def startUpRegistries(my_interface=None):
    # Startup the registries and export them to the GPI, also add jobtree and shareref
    from GangaCore.Runtime.GPIexport import exportToInterface
    if not my_interface:
        import GangaCore.GPI
        my_interface = GangaCore.GPI
    # import default runtime modules

    global _runtime_interface
    _runtime_interface = my_interface

    # bootstrap user-defined runtime modules and enable transient named
    # template registries

    # bootstrap runtime modules
    from GangaCore.GPIDev.Lib.JobTree import TreeError

    for n, k, d in bootstrap():
        # make all repository proxies visible in GPI
        exportToInterface(my_interface, n, k, 'Objects', d)

    # JobTree
    from GangaCore.Core.GangaRepository import getRegistry
    jobtree = getRegistry("jobs").getJobTree()
    exportToInterface(my_interface, 'jobtree', jobtree, 'Objects', 'Logical tree view of the jobs')
    exportToInterface(my_interface, 'TreeError', TreeError, 'Exceptions')

    # ShareRef
    shareref = getRegistry("prep").getShareRef()
    exportToInterface(my_interface, 'shareref', shareref, 'Objects', 'Mechanism for tracking use of shared directory resources')

def removeRegistries():
    ## Remove lingering Objects from the GPI and fully cleanup after the startup

    ## First start with repositories

    import GangaCore.GPI

    from GangaCore.Runtime import Repository_runtime

    to_remove = Repository_runtime.bootstrap_reg_names()
    to_remove.append('jobtree')
    to_remove.append('shareref')

    global _runtime_interface
    for name in to_remove:
        if hasattr(_runtime_interface, name):
            delattr(_runtime_interface, name)

    _runtime_interface = None




from GangaCore.Core.exceptions import GangaException, GangaTypeError
from GangaCore.GPIDev.Base.Objects import GangaObject, Node
from GangaCore.GPIDev.Base.Filters import allComponentFilters
from GangaCore.GPIDev.Base.Proxy import isProxy, addProxy, isType, getProxyAttr, stripProxy, TypeMismatchError, ReadOnlyObjectError, getName
from GangaCore.GPIDev.Base.VPrinter import full_print, summary_print
from GangaCore.GPIDev.Schema.Schema import ComponentItem, Schema, SimpleItem, Version
from GangaCore.GPIDev.Base.Objects import synchronised
from GangaCore.Utility.util import containsGangaObjects, unique
import copy
import sys
from functools import partial

from GangaCore.Utility.logging import getLogger
logger = getLogger(modulename=True)


def makeGangaList(_list, mapfunction=None, parent=None, preparable=False, extra_args=None):
    """Should be used for makeing full gangalists
    Args:
        _list (list): This is the iterable list object which
        mapfunction (function): This is a function used to construct new elements based upon the elements in _list
        parent (GangaObject): This is the object to assign as the parent of the new GangaList (still needed)
        preparable (bool): Is the GangaList preparable?
        extra_args (dict): When defined this contains extra args to pass to mapfunction
    """

    # work with a simple list always
    if isinstance(_list, list):
        _list = _list
    elif isinstance(_list, GangaList):
        _list = getProxyAttr(_list, '_list')
    else:
        _list = [_list]

    if mapfunction is not None:
        if extra_args is None:
            _list = [mapfunction(l) for l in _list]
        else:
            new_mapfunction = partial(mapfunction, extra_args=extra_args)
            _list = [new_mapfunction(l) for l in _list]

    result = GangaList()
    # Subvert tests and modify the ._list here ourselves
    # This is potentially DANGEROUS if proxies aren't correctly stripped
    result._list.extend([stripProxy(l) for l in _list])
    result._is_preparable = preparable
    result._is_a_ref = False

    # set the parent if possible
    if parent is not None:
        result._setParent(parent)

    return result


def stripGangaList(_list):
    """Gets the underlying list of non-proxy objects
    Args:
        _list (list): This is the iterable list of objects which will have their Ganga proxies (if any) stripped
    """
    result = _list
    if isType(_list, GangaList):
        result = getProxyAttr(_list, '_list')
    return result


def makeGangaListByRef(_list, preparable=False):
    """Faster version of makeGangaList. Does not make a copy of _list but use it by reference.
    Args:
        _list (list): List of objects to add to a new Ganga list
        preparable (bool): Is the new object preparable?
    """
    result = GangaList()
    # Subvert tests and modify the ._list here ourselves
    # This is potentially DANGEROUS is proxies aren't correctly stripped
    result._list.extend(_list)
    result._is_a_ref = True
    result._is_preparable = preparable
    return result


def decorateListEntries(entries, typename):
    """
    Generate a small description of the list object
    Args:
        entries (int): Number of entries
        typename (str): Typename of the entries
    """
    return "[%d Entries of type '%s']" % (entries, typename)


class GangaListIter(object):

    """Simple wrapper around the listiterator"""

    __slots__=('it',)

    def __init__(self, it):
        self.it = it

    def next(self):
        # TODO determine if this is correct or needed?
        return addProxy(next(self.it))

    def __iter__(self):
        return self

class GangaList(GangaObject):

    _category = 'internal'
    _exportmethods = ['__add__', '__contains__', '__delitem__', '__delslice__', '__eq__', '__ge__',
                      '__getitem__', '__getslice__', '__gt__', '__iadd__', '__imul__',
                      '__iter__', '__le__', '__len__', '__lt__', '__mul__', '__ne__', '__reversed__', '__radd__', '__rmul__',
                      '__setitem__', '__setslice__', 'append', 'count', 'extend', 'index',
                      'insert', 'pop', 'remove', 'reverse', 'sort', '__hash__', 'get']
    _hidden = 1
    _enable_plugin = 1
    _name = 'GangaList'
    _schema = Schema(Version(1, 0), {'_list': SimpleItem(defvalue=[], doc='The raw list', hidden=1, category='internal'),
                                     '_is_preparable': SimpleItem(defvalue=False, doc='defines if prepare lock is checked', hidden=1),
                                    })
    _enable_config = 1

    _additional_slots = ['_is_a_ref']

    def __init__(self):
        self._is_a_ref = False
        super(GangaList, self).__init__()

    # convenience methods
    @staticmethod
    def is_list(obj):
        """
        This returns a boolean as to if this object is a list or not
        Args:
            obj (object): object to be tested against known list types
        """
        result = (obj is not None) and isType(obj, (GangaList, list, tuple))
        return result

    @staticmethod
    def has_proxy_element(_list):
        """
        Returns if a proxy object has crept into the list
        Args:
            _list (list): Any iterable object
        """
        return all([isProxy(l) for l in _list])

    ## Attempt to prevent raw assignment of _list causing Proxied objects to get inside the GangaList
    def _attribute_filter__set__(self, name, value):
        logger.debug("GangaList filter")
        if name == "_list":
            if self.is_list(value):
                if self.has_proxy_element(value):
                    returnable_list = [stripProxy(l) for l in value]
                    my_parent = self._getParent()
                    for elem in returnable_list:
                        if isinstance(elem, GangaObject):
                            elem._setParent(my_parent)
                        return
                    return returnable_list
                else:
                    return value
            elif self._list is None:
                return None
            else:
                raise GangaException("Attempting to assign a non list item: %s to a GangaList._list!" % str(value))
        else:
            return super(GangaList, self)._attribute_filter__set__(name, value)

    def _on_attribute__set__(self, obj_type, attrib_name):
        if self._is_a_ref is True:
            new_list = []
            for i in self._list:
                if hasattr(i, '_on_attribute__set__'):
                    new_list.append(i._on_attribute__set__(obj_type, attrib_name))
                else:
                    new_list.append(i)

            self._list = new_list
            self._is_a_ref = False
        return self

    def _getParent(self):
        return super(GangaList, self)._getParent()

    def _setParent(self, parent):
        """
        Set the parent of this object and it's children to the given parent
        Args:
            parent (GangaObject): Sets this object as the parent of the list and it's children
        """
        super(GangaList, self)._setParent(parent)
        for elem in self._list:
            if isinstance(elem, GangaObject):
                elem._setParent(parent)

    def get(self, to_match):
        def matching_filter(item):
            if '_list_get__match__' in dir(item):
                return item._list_get__match__(to_match)
            return to_match == item
        return makeGangaListByRef(filter(matching_filter, self._list), preparable=self._is_preparable)

    def _export_get(self, to_match):
        return addProxy(self.get(stripProxy(to_match)))

    def strip_proxy(self, obj, filter=False):
        """Removes proxies and calls shortcut if needed"""

        def applyFilter(obj, item):
            category = item['category']
            this_filter = allComponentFilters[category]
            filter_obj = this_filter(obj, item)
            if filter_obj is None:
                raise TypeMismatchError('%s is not of type %s.' % (str(obj), category))
            return filter_obj

        raw_obj = stripProxy(obj)
        # apply a filter if possible
        if filter is True:
            parent = self._getParent()
            item = self.findSchemaParentSchemaEntry(parent)
            if item and item.isA(ComponentItem):  # only filter ComponentItems
                category = item['category']
                if isinstance(raw_obj, GangaObject):
                    if raw_obj._category != category:
                        raw_obj = applyFilter(raw_obj, item)
                    raw_obj._setParent(parent)
                else:
                    raw_obj = applyFilter(raw_obj, item)
        return raw_obj

    def strip_proxy_list(self, obj_list, filter=False):

        if isType(obj_list, GangaList):
            return getProxyAttr(obj_list, '_list')
        result = [self.strip_proxy(l, filter=filter) for l in obj_list]
        return result

    def getCategory(self):
        """Returns a list of categories for the objects in the list. Returns [] for an empty list."""
        def return_cat(elem):
            if hasattr(elem, '_category'):
                return elem._category
            else:
                return type(elem)
        return unique([return_cat(l) for l in self._list])

    def _readonly(self):
        """
        Return if this object is read-only based upon the Schema
        """
        if self._is_preparable and hasattr(self, '_getParent'):
            if self._getParent()._category == 'applications' and hasattr(self._getParent(), 'is_prepared'):
                from GangaCore.GPIDev.Lib.File.File import ShareDir
                return (isinstance(self._getParent().is_prepared, ShareDir) or super(GangaList, self)._readonly())
        return super(GangaList, self)._readonly()

    def checkReadOnly(self):
        """Puts a hook in to stop mutable access to readonly jobs."""
        if self._readonly():
            raise ReadOnlyObjectError(
                'object %s is readonly and attribute "%s" cannot be modified now' % (repr(self), getName(self)))
        else:
            self._getSessionLock()
            # TODO: BUG: This should only be set _after_ the change has been
            # done! This can lead to data loss!
            self._setDirty()

    # list methods
    # All list methods should be overridden in a way that makes
    # sure that no proxy objects end up in the list, and no
    # unproxied objects make it out.

    def __add__(self, obj_list):
        # Savanah 32342
        if not self.is_list(obj_list):
            raise GangaTypeError('Type %s can not be concatinated to a GangaList' % type(obj_list))

        return makeGangaList(self._list.__add__(self.strip_proxy_list(obj_list, True)), preparable=self._is_preparable)

    def _export___add__(self, obj_list):
        self.checkReadOnly()
        return addProxy(self.__add__(obj_list))

    def __contains__(self, obj):
        return self._list.__contains__(self.strip_proxy(obj))

    def __clone__(self):
        """ clone this object in a similar way to copy """
        # TODO deterine if silently calling __copy__ is more correct
        return makeGangaListByRef(_list=copy.copy(self._list), preparable=self._is_preparable)

    def __copy__(self):
        """Bypass any checking when making the copy"""
        return makeGangaListByRef(_list=copy.copy(self._list), preparable=self._is_preparable)

    def __delitem__(self, obj):
        self._list.__delitem__(self.strip_proxy(obj))

    def _export___delitem__(self, obj):
        self.checkReadOnly()
        self.__delitem__(obj)

    def __delslice__(self, start, end):
        self._list.__delslice__(start, end)

    def _export___delslice__(self, start, end):
        self.checkReadOnly()
        self.__delslice__(start, end)

    def __deepcopy__(self, memo=None):
        """Bypass any checking when making the copy"""
        #logger.info("memo: %s" % str(memo))
        #logger.info("self.len: %s" % str(len(self._list)))
        if self._list != []:
            return makeGangaListByRef(_list=copy.deepcopy(self._list, memo), preparable=self._is_preparable)
        else:
            new_list = GangaList()
            new_list._is_preparable = self._is_preparable
            return new_list

    def __getListToCompare(self, input_list):

        # if the arg isn't a list, just give it back
        if not self.is_list(self.strip_proxy(input_list)):
            return input_list

        # setup up the list correctly
        tmp_list = input_list
        if isType(input_list, GangaList):
            # GangaLists should never contain proxied objects so just return the list
            return stripProxy(input_list)._list
        elif isinstance(input_list, tuple):
            tmp_list = list(input_list)

        # Now return the list after stripping any objects of proxies
        return self.strip_proxy_list(tmp_list)

    def __eq__(self, obj_list):
        if obj_list is self:  # identity check
            return True
        return self._list == self.__getListToCompare(obj_list)

    def __ge__(self, obj_list):
        return self._list.__ge__(self.__getListToCompare(obj_list))

    def __getitem__(self, index):
        return self._list.__getitem__(index)

    def _export___getitem__(self, index):
        return addProxy(self.__getitem__(index))

    def __getslice__(self, start, end):
        return makeGangaList(_list=self._list.__getslice__(start, end), preparable=self._is_preparable)

    def _export___getslice__(self, start, end):
        return addProxy(self.__getslice__(start, end))

    def __gt__(self, obj_list):
        return self._list.__gt__(self.strip_proxy_list(obj_list))

    def __hash__(self):
        logger.info("hash")
        result = 0
        for element in result:
            result ^= hash(element)
        return result

    def __iadd__(self, obj_list):
        self._list.__iadd__(self.strip_proxy_list(obj_list, True))
        return self

    def _export___iadd__(self, obj_list):
        self.checkReadOnly()
        return addProxy(self.__iadd__(obj_list))

    def __imul__(self, number):
        self._list.__imul__(number)
        return self

    def _export___imul__(self, number):
        self.checkReadOnly()
        return addProxy(self.__imul__(number))

    def __iter__(self):
        return self._list.__iter__()

    def _export___iter__(self):
        return GangaListIter(iter(self._list))

    def __le__(self, obj_list):
        return self._list.__le__(self.strip_proxy_list(obj_list))

    def __len__(self):
        return len(self._list)

    def __lt__(self, obj_list):
        return self._list.__lt__(self.strip_proxy_list(obj_list))

    def __mul__(self, number):
        return makeGangaList(self._list.__mul__(number), preparable=self._is_preparable)

    def _export___mul__(self, number):
        return addProxy(self.__mul__(number))

    def __ne__(self, obj_list):
        if obj_list is self:  # identity check
            return False
        result = True
        if self.is_list(obj_list):
            result = self._list.__ne__(self.strip_proxy_list(obj_list))
        return result

    def __reversed__(self):
        """Implements the __reversed__ list method introduced in 2.4"""
        return reversed(self._list)

    def _export___reversed__(self):
        return GangaListIter(self.__reversed__())

    def __radd__(self, obj):
        return obj + self._list

    def _export___radd__(self, obj):
        # return the proxied objects
        cp = [i for i in self._export___iter__()]
        return obj + cp

    def __rmul__(self, number):
        return makeGangaList(self._list.__rmul__(number), preparable=self._is_preparable)

    def _export___rmul__(self, number):
        return addProxy(self.__rmul__(number))

    def __setitem__(self, index, obj):
        self._list.__setitem__(index, self.strip_proxy(obj, True))

    def _export___setitem__(self, index, obj):
        self.checkReadOnly()
        self.__setitem__(index, obj)

    def __setslice__(self, start, end, obj_list):
        self._list.__setslice__(start, end, self.strip_proxy_list(obj_list, True))

    def _export___setslice__(self, start, end, obj_list):
        self.checkReadOnly()
        self.__setslice__(start, end, obj_list)

    def __repr__(self):
        #logger.info("__repr__")
        #return self.toString()
        #import traceback
        #traceback.print_stack()
        containsObj = False
        for elem in self._list:
            if isinstance(elem, GangaObject):
                containsObj = True
                break
        if containsObj is False:
            return self.toString()
        else:
            return str("<GangaList at: %s>" % str(hex(abs(id(self)))))
        #return str("<GangaList at: %s>" % str(hex(abs(id(self)))))

    def __str__(self):
        #logger.info("__str__")
        return self.toString()

    def append(self, obj, my_filter=True):
        if isType(obj, GangaList):
            stripped_o = stripProxy(obj)
            stripped_o._setParent(self._getParent())
            self._list.append(stripped_o)
            return
        elem = self.strip_proxy(obj, my_filter)
        list_objs = (list, tuple)
        if isType(elem, GangaObject):
            stripped_e = stripProxy(elem)
            stripped_e._setParent(self._getParent())
            self._list.append(stripped_e)
        elif isType(elem, list_objs):
            new_list = []
            def my_append(_obj):
                if isType(_obj, GangaObject):
                    stripped_o = stripProxy(_obj)
                    stripped_o._setParent(self._getParent())
                    return stripped_o
                else:
                    return _obj
            self._list.append([my_append(l) for l in elem])
        else:
            self._list.append(elem)

    def _export_append(self, obj):
        self.checkReadOnly()
        self.append(obj)

    def count(self, obj):
        return self._list.count(self.strip_proxy(obj))

    def extend(self, ittr):
        for i in ittr:
            self.append(i)

    def _export_extend(self, ittr):
        self.checkReadOnly()
        self.extend(ittr)

    def index(self, obj):
        return self._list.index(self.strip_proxy(obj))

    def insert(self, index, obj):
        if isType(obj, GangaObject):
            stripProxy(obj)._setParent(stripProxy(self)._getParent())
        self._list.insert(index, self.strip_proxy(obj, True))

    def _export_insert(self, index, obj):
        self.checkReadOnly()
        self.insert(index, obj)

    def pop(self, index=-1):
        return self._list.pop(index)

    def _export_pop(self, index=-1):
        self.checkReadOnly()
        return addProxy(self.pop(index))

    def remove(self, obj):
        """
        Remove a given object from a list
        Args:
            obj (unknown): Remove this object from the list if it exists
        """
        self._list.remove(self.strip_proxy(obj))

    def _export_remove(self, obj):
        """
        Remove an object from a list after checking if it's read-only
        Args:
            obj (unknown): Remove this object from the list if it exists
        """
        self.checkReadOnly()
        self.remove(obj)

    def reverse(self):
        self._list.reverse()

    def _export_reverse(self):
        self.checkReadOnly()
        self.reverse()

    def sort(self, cmpfunc=None):
        # TODO: Should comparitor have access to unproxied objects?
        self._list.sort(cmpfunc)

    def _export_sort(self, cmpfunc=None):
        """
        Args:
            compfunc (function): Function to be used in sorting list elements
        """
        self.checkReadOnly()
        self.sort(cmpfunc)

    # now some more ganga specific methods
    def findSchemaParentSchemaEntry(self, parent):
        """Finds the schema entry for this GangaList
        Args:
            parent (GangaObject): This is the parent object which controls this object
        """
        result = None
        if parent and parent._schema:
            for k, v in parent._schema.allItems():
                if getattr(parent, k) is self:
                    result = v
                    break
        return result

    def printSummaryTree(self, level=0, verbosity_level=0, whitespace_marker='', out=sys.stdout, selection='', interactive=False):
        """
        This funtion displays a summary of the contents of this file.
        (Docs from Gaga.GPIDev.Base.Objects # TODO determine how mch of this may be duplicated from there)
        Args:
            level (int): the hierachy level we are currently at in the object tree.
            verbosity_level (int): How verbose the print should be. Currently this is always 0.
            whitespace_marker (str): If printing on multiple lines, this allows the default indentation to be replicated.
                                     The first line should never use this, as the substitution is 'name = %s' % printSummaryTree()
            out (stream): An output stream to print to. The last line of output should be printed without a newline.'
            selection: See VPrinter for an explaintion of this.
            interactive (bool): Is this printing code being called from the interactive IPython prompt?
        """
        parent = self._getParent()
        schema_entry = self.findSchemaParentSchemaEntry(parent)

        if parent is None:
            full_print(self, out)
            return

        if schema_entry:
            self_len = len(self)
            print_summary = schema_entry['summary_print']
            maxLen = schema_entry['summary_sequence_maxlen']

            if print_summary:
                fp = getattr(parent, print_summary)
                str_val = fp(self._list, verbosity_level)
                out.write(str_val)
                return

            if (maxLen != -1) and (self_len > maxLen):
                out.write(decorateListEntries(self_len, getName(type(self[0]))))
                return
            else:
                summary_print(self, out)
                return

        out.write(str(self._list))
        return

    def toString(self):
        """Returns a simple str of the _list."""
        returnable_str = "["
        for element in self._list:
            if isType(element, GangaObject):
                returnable_str += repr(stripProxy(element))
            else:
                returnable_str += "'"
                returnable_str += str(stripProxy(element))
                returnable_str += "'"
            returnable_str += ", "
        returnable_str += "]"
        return returnable_str

    @synchronised
    def accept(self, visitor):
        """
        accept a visitor pattern - overrides GangaObject because we need to process _list as a ComponentItem in this
        case to allow save/load of nested lists to work.
        Could just get away with SimpleItem checks here but have included Shared and Component just in case these
        are added in the future

        # TODO investigate if it's possible to call GangaObject.accept for 90% of this functionality rather than duplicating it here

        This accepts a VPrinter or VStreamer class instance visitor which is used to display the contents of the object according the Schema
        Args:
            visitor (VPrinter, VStreamer): An instance which will produce a display of this contents of this object
        """
        visitor.nodeBegin(self)

        for (name, item) in self._schema.simpleItems():
            if name == "_list":
                visitor.componentAttribute(self, name, getattr(self, name), 1)
            elif item['visitable']:
                visitor.simpleAttribute(self, name, getattr(self, name), item['sequence'])

        for (name, item) in self._schema.sharedItems():
            if item['visitable']:
                visitor.sharedAttribute(self, name, getattr(self, name), item['sequence'])

        for (name, item) in self._schema.componentItems():
            if item['visitable']:
                visitor.componentAttribute(self, name, getattr(self, name), item['sequence'])

        visitor.nodeEnd(self)

    def _setFlushed(self):
        """Set flushed like the Node but do the _list by hand to avoid problems"""
        self._dirty = False
        for elem in self._list:
            if isinstance(elem, Node):
                elem._setFlushed()
        super(GangaList, self)._setFlushed()

# export to GPI moved to the Runtime bootstrap




#!/usr/bin/env python

import sys
import os.path

doc = sys.argv[1]
filename = 'html/%s/%s.html'%(doc,doc)

path = os.path.abspath('.')+'/ganga-%s/'%(doc)

lines =  file(filename).readlines()
of = file(filename,'w')
for l in lines:
    l = l.replace(path,'')
    print >>of, l,

